"""Power Supply Data."""

import copy as _copy


from ..search import PSSearch as _PSSearch

from .csdev import get_ps_propty_database as _get_ps_propty_database


class PSData:
    """PSData Class.

    Class objects that group power supply data.
    """

    def __init__(self, psname):
        """Init method."""
        if psname not in _PSSearch.get_psnames(filter_auxps=False):
            raise ValueError(psname + ' is not a valid power supply name.')
        self._psname = psname
        self._psmodel = _PSSearch.conv_psname_2_psmodel(self._psname)
        self._pstype = _PSSearch.conv_psname_2_pstype(self._psname)
        self._magfunc = _PSSearch.conv_pstype_2_magfunc(self._pstype)
        if self._magfunc != 'none':
            self._polarity = _PSSearch.conv_pstype_2_polarity(self._pstype)
            self._splims = _PSSearch.conv_pstype_2_splims(self._pstype)
            self._splims_unit = _PSSearch.get_splims_unit(self._psmodel)
            self._excdata = _PSSearch.conv_psname_2_excdata(self._psname)
        self._propty_database = \
            _get_ps_propty_database(self._psmodel, self._pstype)

    @property
    def psname(self):
        """Power supply name."""
        return self._psname

    @property
    def pstype(self):
        """Power supply type."""
        return self._pstype

    @property
    def psmodel(self):
        """Power supply model."""
        return self._psmodel

    @property
    def polarity(self):
        """Power supply polarity.

        It can be either 'monopolar' or 'bipolar'.
        """
        return self._polarity

    @property
    def magfunc(self):
        """Magnetic function the power supply excites."""
        return self._magfunc

    @property
    def splims(self):
        """Setpoint limits of power supply."""
        if self._splims is None:
            return None
        else:
            return self._splims.copy()

    @property
    def splims_unit(self):
        """Return unit of power supply setpoint limits."""
        return self._splims_unit

    @property
    def splims_labels(self):
        """EPICS labels of power supply setpoint limits."""
        return sorted(self._splims.keys())

    @property
    def propty_database(self):
        """Power supply properties database."""
        return _copy.deepcopy(self._propty_database)

    @property
    def excdata(self):
        """Excitation data associated with the power supply."""
        return self._excdata

    def __str__(self):
        """__str__ method."""
        rst = ''
        rst += 'psname      : ' + str(self.psname)
        rst += '\n' + 'pstype      : ' + str(self.pstype)
        rst += '\n' + 'polarity    : ' + str(self.polarity)
        rst += '\n' + 'magfunc     : ' + str(self.magfunc)
        rst += '\n' + 'splims_unit : ' + str(self.splims_unit)

        if self.splims is None:
            rst += '\n' + 'splims      : ' + str(None)
        else:
            rst += '\n' + 'splims      : ' + \
                'DRVH:{0:+09.3f} DRVL:{1:+09.3f}'.format(
                    self.splims['DRVH'], self.splims['DRVL'])
            rst += '\n' + '              ' + \
                'HIHI:{0:+09.3f} LOLO:{1:+09.3f}'.format(
                    self.splims['HIHI'],    self.splims['LOLO'])
            rst += '\n' + '              ' + \
                'HIGH:{0:+09.3f} LOW :{1:+09.3f}'.format(
                    self.splims['HIGH'], self.splims['LOW'])
            rst += '\n' + '              ' + \
                'HOPR:{0:+09.3f} LOPR:{1:+09.3f}'.format(
                    self.splims['HOPR'], self.splims['LOPR'])

        return rst



"""PS Search module."""

import copy as _copy
from threading import Lock as _Lock

from .. import util as _util
from .. import clientweb as _web
from ..namesys import Filter as _Filter
from ..namesys import SiriusPVName as _SiriusPVName
from ..pwrsupply.siggen import SigGenFactory as _SigGenFactory
from ..magnet.excdata import ExcitationData as _ExcitationData


class PSSearch:
    """PS and PU Power Supply Search Class."""

    # auxiliary BO dipole bsmp devices not to be considered
    # power supplies from standpoint of high level appplications
    _bo_dip_auxps = \
        set('BO-Fam:PS-B-' + idx for idx in
            ('1a', '1b', '1c', '2a', '2b', '2c'))

    _splims_labels = list()
    _splims_ps_unit = list()
    _splims_pu_unit = list()
    _psnames_list = list()

    _pstype_dict = dict()
    _pstype_2_psnames_dict = dict()
    _pstype_2_splims_dict = dict()
    _pstype_2_excdat_dict = dict()
    _psname_2_psmodel_dict = dict()
    _psmodel_2_psname_dict = dict()
    _psname_2_siggen_dict = dict()
    _bbbname_2_bsmps_dict = dict()
    _bsmps_2_bbbname_dict = dict()
    _bbbname_2_udc_dict = dict()
    _bbbname_2_freq_dict = dict()
    _udc_2_bbbname_dict = dict()
    _udc_2_bsmp_dict = dict()
    _bsmp_2_udc_dict = dict()
    _ps_2_dclink_dict = dict()
    _dclink_2_ps_dict = dict()

    _lock = _Lock()

    _linac_ps_sinap2sirius = {
        'LA-CN:H1DPPS-1': 'LI-01:PS-Spect',
        'LA-CN:H1FQPS-1': 'LI-Fam:PS-QF1',
        'LA-CN:H1FQPS-2': 'LI-Fam:PS-QF2',
        'LA-CN:H1FQPS-3': 'LI-01:PS-QF3',
        'LA-CN:H1DQPS-1': 'LI-01:PS-QD1',
        'LA-CN:H1DQPS-2': 'LI-01:PS-QD2',
        'LA-CN:H1SCPS-1': 'LI-01:PS-CV-1',
        'LA-CN:H1SCPS-2': 'LI-01:PS-CH-1',
        'LA-CN:H1SCPS-3': 'LI-01:PS-CV-2',
        'LA-CN:H1SCPS-4': 'LI-01:PS-CH-2',
        'LA-CN:H1LCPS-1': 'LI-01:PS-CV-3',
        'LA-CN:H1LCPS-2': 'LI-01:PS-CH-3',
        'LA-CN:H1LCPS-3': 'LI-01:PS-CV-4',
        'LA-CN:H1LCPS-4': 'LI-01:PS-CH-4',
        'LA-CN:H1LCPS-5': 'LI-01:PS-CV-5',
        'LA-CN:H1LCPS-6': 'LI-01:PS-CH-5',
        'LA-CN:H1LCPS-7': 'LI-01:PS-CV-6',
        'LA-CN:H1LCPS-8': 'LI-01:PS-CH-6',
        'LA-CN:H1LCPS-9': 'LI-01:PS-CV-7',
        'LA-CN:H1LCPS-10': 'LI-01:PS-CH-7',
        'LA-CN:H1RCPS-1': 'LI-01:PS-LensRev',
        'LA-CN:H1MLPS-1': 'LI-01:PS-Lens-1',
        'LA-CN:H1MLPS-2': 'LI-01:PS-Lens-2',
        'LA-CN:H1MLPS-3': 'LI-01:PS-Lens-3',
        'LA-CN:H1MLPS-4': 'LI-01:PS-Lens-4',
        'LA-CN:H1SLPS-1': 'LI-01:PS-Slnd-1',
        'LA-CN:H1SLPS-2': 'LI-01:PS-Slnd-2',
        'LA-CN:H1SLPS-3': 'LI-01:PS-Slnd-3',
        'LA-CN:H1SLPS-4': 'LI-01:PS-Slnd-4',
        'LA-CN:H1SLPS-5': 'LI-01:PS-Slnd-5',
        'LA-CN:H1SLPS-6': 'LI-01:PS-Slnd-6',
        'LA-CN:H1SLPS-7': 'LI-01:PS-Slnd-7',
        'LA-CN:H1SLPS-8': 'LI-01:PS-Slnd-8',
        'LA-CN:H1SLPS-9': 'LI-01:PS-Slnd-9',
        'LA-CN:H1SLPS-10': 'LI-01:PS-Slnd-10',
        'LA-CN:H1SLPS-11': 'LI-01:PS-Slnd-11',
        'LA-CN:H1SLPS-12': 'LI-01:PS-Slnd-12',
        'LA-CN:H1SLPS-13': 'LI-01:PS-Slnd-13',
        'LA-CN:H1SLPS-14': 'LI-Fam:PS-Slnd-14',
        'LA-CN:H1SLPS-15': 'LI-Fam:PS-Slnd-15',
        'LA-CN:H1SLPS-16': 'LI-Fam:PS-Slnd-16',
        'LA-CN:H1SLPS-17': 'LI-Fam:PS-Slnd-17',
        'LA-CN:H1SLPS-18': 'LI-Fam:PS-Slnd-18',
        'LA-CN:H1SLPS-19': 'LI-Fam:PS-Slnd-19',
        'LA-CN:H1SLPS-20': 'LI-Fam:PS-Slnd-20',
        'LA-CN:H1SLPS-21': 'LI-Fam:PS-Slnd-21',
    }

    @staticmethod
    def get_psnames(filters=None, filter_auxps=True):
        """Return a sorted and filtered list of all power supply names."""
        PSSearch._reload_pstype_2_psnames_dict()
        psnames = _Filter.process_filters(
            PSSearch._psnames_list, filters=filters)
        if filter_auxps:
            psnames = set(psnames) - PSSearch._bo_dip_auxps
        return sorted(psnames)

    @staticmethod
    def get_psnicknames(names=None, filters=None):
        """Return a list with Magnet nicknames."""
        if not names:
            names = PSSearch.get_psnames(filters=filters)
        nicknames = len(names)*['']
        for i, pss in enumerate(names):
            nicknames[i] = pss.sub + ('-' + pss.idx if pss.idx else '')
        return nicknames

    @staticmethod
    def get_pstype_names():
        """Return sorted list of power supply types."""
        with PSSearch._lock:
            PSSearch._reload_pstype_dict()
        return sorted(set(PSSearch._pstype_dict.keys()))

    @staticmethod
    def get_psmodel_names():
        """Return sorted list of power supply models."""
        PSSearch._reload_psname_2_psmodel_dict()
        return sorted(set(PSSearch._psmodel_2_psname_dict.keys()))

    @staticmethod
    def get_bbbnames(filters=None):
        """Return a sorted and filtered list of all beaglebone names."""
        PSSearch._reload_bbb_2_bsmps_dict()
        bbname_list = tuple(PSSearch._bbbname_2_bsmps_dict.keys())
        return sorted(_Filter.process_filters(bbname_list, filters=filters))

    @staticmethod
    def get_splims(pstype, label):
        """Return setpoint limit corresponding to given label.

        The label can be either epics' or pcaspy's.
        """
        PSSearch._reload_pstype_2_splims_dict()
        if label in PSSearch._splims_labels:
            return PSSearch._pstype_2_splims_dict[pstype][label]
        else:
            label = _util.conv_splims_labels(label)
            return PSSearch._pstype_2_splims_dict[pstype][label]

    @staticmethod
    def get_pstype_dict():
        """Return pstype dictionary.

        With key,value pairs of power supply types and corresponding
        (polarities,mag_function).
        """
        with PSSearch._lock:
            PSSearch._reload_pstype_dict()
        return _copy.deepcopy(PSSearch._pstype_dict)

    @staticmethod
    def get_bbbname_dict():
        """Return bbbname dictionary.

        With key,value pairs of bbbname and corresponding power supplies.
        """
        PSSearch._reload_bbb_2_bsmps_dict()
        return _copy.deepcopy(PSSearch._bbbname_2_bsmps_dict)

    @staticmethod
    def get_udc_dict():
        """Return udc dictionary.

        With key,value pairs of udc and corresponding power supplies.
        """
        with PSSearch._lock:
            PSSearch._reload_udc_2_bsmp_dict()
        return _copy.deepcopy(PSSearch._udc_2_bsmp_dict)

    @staticmethod
    def get_polarities():
        """Return sorted list of power supply polarities."""
        with PSSearch._lock:
            PSSearch._reload_pstype_dict()
        pol = [datum[0] for datum in PSSearch._pstype_dict.values()]
        return sorted(set(pol))

    @staticmethod
    def conv_psname_2_pstype(psname):
        """Return the power supply type of a given power supply name."""
        PSSearch._reload_pstype_2_psnames_dict()
        for pstype, psnames in PSSearch._pstype_2_psnames_dict.items():
            if psname in psnames:
                return pstype
        raise KeyError('Invalid psname "' + psname + '"!')

    @staticmethod
    def conv_psname_2_splims(psname):
        """Convert maname powersupply to a dict with its setpoint limits."""
        if psname is None:
            return None
        pstype = PSSearch.conv_psname_2_pstype(psname)
        splims = PSSearch.conv_pstype_2_splims(pstype)
        return _copy.deepcopy(splims)

    @staticmethod
    def conv_pstype_2_polarity(pstype):
        """Return polarity of a given power supply type."""
        with PSSearch._lock:
            PSSearch._reload_pstype_dict()
        return PSSearch._pstype_dict[pstype][0]

    @staticmethod
    def conv_pstype_2_magfunc(pstype):
        """Return magnetic function of a given power supply type."""
        with PSSearch._lock:
            PSSearch._reload_pstype_dict()
        for key, value in PSSearch._pstype_dict.items():
            if key == pstype:
                return value[1]
        raise KeyError('Invalid pstype "'+pstype+'"!')

    @staticmethod
    def conv_pstype_2_splims(pstype):
        """Convert pstype to splims."""
        if pstype is None:
            return None
        PSSearch._reload_pstype_2_splims_dict()
        return _copy.deepcopy(PSSearch._pstype_2_splims_dict[pstype])

    @staticmethod
    def conv_psname_2_magfunc(psname):
        """Return magnetic function of a given power supply."""
        pstype = PSSearch.conv_psname_2_pstype(psname)
        psfunc = PSSearch.conv_pstype_2_magfunc(pstype)
        return psfunc

    @staticmethod
    def conv_psname_2_excdata(psname):
        """Convert psname to excdata."""
        pstype = PSSearch.conv_psname_2_pstype(psname)
        PSSearch._reload_pstype_2_excdat_dict(pstype)
        return PSSearch._pstype_2_excdat_dict[pstype]

    @staticmethod
    def conv_psname_2_psmodel(psname):
        """Convert psname to psmodel."""
        PSSearch._reload_psname_2_psmodel_dict()
        return PSSearch._psname_2_psmodel_dict[psname]

    @staticmethod
    def conv_psmodel_2_psname(psmodel):
        """Convert psmodel to psname."""
        PSSearch._reload_psname_2_psmodel_dict()
        return PSSearch._psmodel_2_psname_dict[psmodel]

    @staticmethod
    def conv_psname_2_siggenconf(psname):
        """Convert psname to corresponding SigGenConf object."""
        PSSearch._reload_psname_2_siggen_dict()
        return PSSearch._psname_2_siggen_dict[psname]

    @staticmethod
    def conv_psname_2_bbbname(psname):
        """Given psname return the bbb name."""
        udc = PSSearch.conv_psname_2_udc(psname)
        return PSSearch.conv_udc_2_bbbname(udc)

    @staticmethod
    def conv_bbbname_2_psnames(bbbname):
        """Given bbb name return the psnames."""
        return PSSearch.conv_bbbname_2_bsmps(bbbname)

    @staticmethod
    def conv_bbbname_2_bsmps(bbbname):
        """Given bbb name return bsmps devices."""
        PSSearch._reload_bbb_2_bsmps_dict()
        return PSSearch._bbbname_2_bsmps_dict[bbbname]

    @staticmethod
    def conv_bbbname_2_freq(bbbname):
        """Given bbb name return PRU sync off and on update frequencies."""
        PSSearch._reload_bbb_2_freq_dict()
        return PSSearch._bbbname_2_freq_dict[bbbname]

    @staticmethod
    def conv_bbbname_2_udc(bbbname):
        """Return UDCs connected with a beaglebone."""
        with PSSearch._lock:
            PSSearch._reload_bbb_2_udc_dict()
        return PSSearch._bbbname_2_udc_dict[bbbname]

    @staticmethod
    def conv_udc_2_bbbname(udc):
        """Return beaglebone name connected to a UDC."""
        with PSSearch._lock:
            PSSearch._reload_bbb_2_udc_dict()
        return PSSearch._udc_2_bbbname_dict[udc]

    @staticmethod
    def conv_udc_2_bsmps(udc):
        """Return BSMP devices associated with a given UDC."""
        with PSSearch._lock:
            PSSearch._reload_udc_2_bsmp_dict()
        return PSSearch._udc_2_bsmp_dict[udc]

    @staticmethod
    def conv_psname_2_udc(psname):
        """Return UDC associated with a power supply."""
        with PSSearch._lock:
            PSSearch._reload_udc_2_bsmp_dict()
        return PSSearch._bsmp_2_udc_dict[psname]

    @staticmethod
    def conv_psname_2_dclink(psname, filter_auxps=True):
        """Return DCLink associated with a power supply."""
        # NOTE: lock is being used within private method.
        dict_aux = PSSearch.get_psname_2_dclink_dict(filter_auxps)
        return dict_aux[psname]

    @staticmethod
    def conv_dclink_2_psname(dclink, filter_auxps=True):
        """Return PS associated with a DCLink."""
        # NOTE: lock is being used within private method.
        dict_aux = PSSearch.get_dclink_2_psname_dict(filter_auxps)
        return dict_aux[dclink]

    @staticmethod
    def get_linac_ps_sinap2sirius_dict():
        """Return PS name convertion dict."""
        return PSSearch._linac_ps_sinap2sirius

    @staticmethod
    def get_linac_ps_sirius2sinap_dict():
        """Return PS name convertion dict."""
        dic = PSSearch._linac_ps_sinap2sirius
        return {v: k for k, v in dic.items()}

    @staticmethod
    def get_pstype_2_psnames_dict():
        """Return dictionary of power supply type and power supply names."""
        PSSearch._reload_pstype_2_psnames_dict()
        return _copy.deepcopy(PSSearch._pstype_2_psnames_dict)

    @staticmethod
    def get_pstype_2_splims_dict():
        """Return a dictionary of power supply type and setpoint limits."""
        PSSearch._reload_pstype_2_splims_dict()
        return _copy.deepcopy(PSSearch._pstype_2_splims_dict)

    @staticmethod
    def get_splims_unit(psmodel):
        """Return SP limits unit."""
        PSSearch._reload_pstype_2_splims_dict()
        if psmodel in ('FBP', 'FBP_DCLink',
                       'FAC_DCDC', 'FAC_2S_DCDC', 'FAC_2S_ACDC',
                       'FAC_2P4S_DCDC', 'FAC_2P4S_ACDC',
                       'FAP', 'FAP_2P2S', 'FAP_4P',
                       'Commercial', 'LINAC_PS', 'FOFB_PS'):
            return PSSearch._splims_ps_unit
        elif psmodel in ('APU', ):
            return 'mm'
        elif psmodel in ('FP_SEPT', 'FP_KCKR', 'FP_PINGER'):
            return PSSearch._splims_pu_unit
        else:
            raise ValueError(psmodel)

    @staticmethod
    def get_splims_labels():
        """Return labels in SP limits dictionary."""
        PSSearch._reload_pstype_2_splims_dict()
        return PSSearch._splims_labels

    @staticmethod
    def get_psname_2_dclink_dict(filter_auxps=True):
        """Return a dictionary of power supply and DCLinks."""
        PSSearch._reload_ps_2_dclink_dict()
        dict_aux = _copy.deepcopy(PSSearch._ps_2_dclink_dict)
        if filter_auxps:
            for psn in PSSearch._bo_dip_auxps:
                dict_aux.pop(psn)
        return dict_aux

    @staticmethod
    def get_dclink_2_psname_dict(filter_auxps=True):
        """Return a dictionary of DCLink and power supplies."""
        PSSearch._reload_ps_2_dclink_dict()
        dict_aux = _copy.deepcopy(PSSearch._dclink_2_ps_dict)
        if filter_auxps:
            for dcl, list_aux in PSSearch._dclink_2_ps_dict.items():
                for psn in PSSearch._bo_dip_auxps:
                    if psn in list_aux:
                        dict_aux[dcl].remove(psn)
        return dict_aux

    # --- private methods ---

    @staticmethod
    def _reload_pstype_dict():
        """Reload power supply type dictionary from web server."""
        if PSSearch._pstype_dict:
            return
        if not _web.server_online():
            raise Exception('could not read pstypes from web server!')
        text = _web.ps_pstypes_names_read()
        data, _ = _util.read_text_data(text)
        pstype_dict = dict()
        for datum in data:
            name, polarity, magfunc = datum[0], datum[1], datum[2]
            pstype_dict[name] = (polarity, magfunc)
        PSSearch._pstype_dict = pstype_dict

    @staticmethod
    def _reload_pstype_2_psnames_dict():
        """Reload power supply type to power supply names dictionary."""
        with PSSearch._lock:
            if PSSearch._pstype_2_psnames_dict:
                return
            PSSearch._reload_pstype_dict()
            pstypes = sorted(set(PSSearch._pstype_dict.keys()))
            pstype_2_psnames_dict = dict()
            psnames_list = list()
            for pstype in pstypes:
                text = _web.ps_pstype_data_read(pstype + '.txt')
                data, _ = _util.read_text_data(text)
                psnames = [_SiriusPVName(datum[0]) for datum in data]
                pstype_2_psnames_dict[pstype] = psnames
                psnames_list += psnames
            PSSearch._pstype_2_psnames_dict = pstype_2_psnames_dict
            PSSearch._psnames_list = sorted(psnames_list)

    @staticmethod
    def _reload_pstype_2_splims_dict():
        """Reload pstype to splims dictionary."""
        with PSSearch._lock:
            if PSSearch._pstype_2_splims_dict:
                return
            if not _web.server_online():
                raise Exception(
                    'could not read setpoint limits from web server!')
            # ps data
            text = _web.ps_pstype_setpoint_limits()
            ps_data, ps_param_dict = _util.read_text_data(text)
            # pu data
            text = _web.pu_pstype_setpoint_limits()
            pu_data, pu_param_dict = _util.read_text_data(text)

            # checks consistency between PS and PU static tables.
            # this should be implemented elsewhere, not in PSSearch!
            if sorted(ps_param_dict['power_supply_type']) != \
                    sorted(pu_param_dict['power_supply_type']):
                raise ValueError(('Inconsistent limit labels between PS '
                                  'and PU static tables!'))

            PSSearch._splims_ps_unit = ps_param_dict['unit']
            PSSearch._splims_pu_unit = pu_param_dict['unit']
            PSSearch._splims_labels = ps_param_dict['power_supply_type']

            pstype_2_splims_dict = dict()
            data = ps_data + pu_data
            for datum in data:
                pstype, *lims = datum
                pstype_2_splims_dict[pstype] = \
                    {PSSearch._splims_labels[i]:
                     float(lims[i]) for i in range(len(lims))}
            PSSearch._pstype_2_splims_dict = pstype_2_splims_dict

    @staticmethod
    def _reload_pstype_2_excdat_dict(pstype):
        """Load power supply excitation data."""
        with PSSearch._lock:
            if pstype in PSSearch._pstype_2_excdat_dict:
                return
            if not _web.server_online():
                raise Exception(
                    'could not read "' + str(pstype) + '" from web server!')
            PSSearch._pstype_2_excdat_dict[pstype] = \
                _ExcitationData(filename_web=pstype + '.txt')

    @staticmethod
    def _reload_psname_2_psmodel_dict():
        """Load psmodels by psname to a dict."""
        with PSSearch._lock:
            if PSSearch._psname_2_psmodel_dict:
                return
            if not _web.server_online():
                raise Exception('could not read psmodels from web server')
            ps_data, _ = _util.read_text_data(_web.ps_psmodels_read())
            pu_data, _ = _util.read_text_data(_web.pu_psmodels_read())
            data = ps_data + pu_data
            psname_2_psmodel_dict = dict()
            psmodel_2_psname_dict = dict()
            for datum in data:
                psname, psmodel = datum
                psname_2_psmodel_dict[psname] = psmodel
                if psmodel not in psmodel_2_psname_dict:
                    psmodel_2_psname_dict[psmodel] = list()
                psmodel_2_psname_dict[psmodel].append(psname)
            PSSearch._psname_2_psmodel_dict = psname_2_psmodel_dict
            PSSearch._psmodel_2_psname_dict = psmodel_2_psname_dict

    @staticmethod
    def _reload_psname_2_siggen_dict():
        """Load siggen config by psname to a dict."""
        with PSSearch._lock:
            if PSSearch._psname_2_siggen_dict:
                return
            if not _web.server_online():
                raise Exception(
                    'could not read siggen config from web server')
            text = _web.ps_siggen_configuration_read()
            data, _ = _util.read_text_data(text)
            psname_2_siggen_dict = dict()
            for datum in data:
                psname, *siggen_data = datum
                signal = _SigGenFactory.create(data=siggen_data)
                psname_2_siggen_dict[psname] = signal
            PSSearch._psname_2_siggen_dict = psname_2_siggen_dict

    @staticmethod
    def _reload_bbb_2_bsmps_dict():
        """Test."""
        with PSSearch._lock:
            if PSSearch._bbbname_2_bsmps_dict:
                return
            PSSearch._reload_bbb_2_udc_dict()
            PSSearch._reload_udc_2_bsmp_dict()

            bbbname_2_bsmps_dict = dict()
            for bbb, udcs in PSSearch._bbbname_2_udc_dict.items():
                bsmps = []
                for udc in udcs:
                    bsmps.extend(PSSearch._udc_2_bsmp_dict[udc])
                bbbname_2_bsmps_dict[bbb] = bsmps
            PSSearch._bbbname_2_bsmps_dict = bbbname_2_bsmps_dict

            bsmps_2_bbbname_dict = dict()
            for bbb, bsmps in PSSearch._bbbname_2_bsmps_dict.items():
                bsmps_2_bbbname_dict.update({x: bbb for x in bsmps})
            PSSearch._bsmps_2_bbbname_dict = bsmps_2_bbbname_dict

    @staticmethod
    def _reload_bbb_2_freq_dict():
        with PSSearch._lock:
            if PSSearch._bbbname_2_freq_dict:
                return
            if not _web.server_online():
                raise Exception(
                    'could not read BBB frequency map from web server')
            data, _ = _util.read_text_data(_web.beaglebone_freq_mapping())
            bbbname_2_freq_dict = dict()
            for line in data:
                bbbname, sync_off_freq = line
                bbbname_2_freq_dict[bbbname] = float(sync_off_freq)
            PSSearch._bbbname_2_freq_dict = bbbname_2_freq_dict

    @staticmethod
    def _reload_bbb_2_udc_dict():
        if PSSearch._bbbname_2_udc_dict:
            return
        if not _web.server_online():
            raise Exception('could not read BBB to UDC map from web server')
        data, _ = _util.read_text_data(_web.bbb_udc_mapping())
        bbbname_2_udc_dict = dict()
        udc_2_bbbname_dict = dict()
        for line in data:
            bbbname, *udcs = line
            bbbname_2_udc_dict[bbbname] = udcs
            for udc in udcs:
                udc_2_bbbname_dict[udc] = bbbname
        PSSearch._bbbname_2_udc_dict = bbbname_2_udc_dict
        PSSearch._udc_2_bbbname_dict = udc_2_bbbname_dict

    @staticmethod
    def _reload_udc_2_bsmp_dict():
        if PSSearch._udc_2_bsmp_dict:
            return
        if not _web.server_online():
            raise Exception('could not read UDC to BSMP map from web server')
        data, _ = _util.read_text_data(_web.udc_ps_mapping())
        udc_2_bsmp_dict = dict()
        bsmp_2_udc_dict = dict()
        for line in data:
            udc, *bsmps = line
            udc_2_bsmp_dict[udc] = list()
            for i in range(len(bsmps)//2):
                bsmp = bsmps[2*i+0], int(bsmps[2*i+1])
                udc_2_bsmp_dict[udc].append(bsmp)
                bsmp_2_udc_dict[bsmp[0]] = udc
        PSSearch._udc_2_bsmp_dict = udc_2_bsmp_dict
        PSSearch._bsmp_2_udc_dict = bsmp_2_udc_dict

    @staticmethod
    def _reload_ps_2_dclink_dict():
        with PSSearch._lock:
            if PSSearch._ps_2_dclink_dict:
                return
            if not _web.server_online():
                raise Exception(
                    'could not read BSMP to DCLink map from web server')
            data, _ = _util.read_text_data(_web.bsmp_dclink_mapping())
            ps_2_dclink_dict = dict()
            dclink_2_ps_dict = dict()
            for line in data:
                dclinks = line[1:]
                if dclinks[0] == 'None':
                    ps_2_dclink_dict[line[0]] = None
                else:
                    ps_2_dclink_dict[line[0]] = dclinks
                    for dcl in dclinks:
                        if dcl not in dclink_2_ps_dict:
                            dclink_2_ps_dict[dcl] = list()
                        dclink_2_ps_dict[dcl].append(line[0])

            PSSearch._ps_2_dclink_dict = ps_2_dclink_dict
            PSSearch._dclink_2_ps_dict = dclink_2_ps_dict



"""Oscilloscopes and scope psignals."""

class Scopes:
    """Keysight oscilloscopes names and IPs."""

    AS_DI_FCTDIG = '10.128.150.78'
    AS_DI_FPMDIG = '10.128.150.77'
    LI_DI_ICTOSC = '10.128.1.52'
    TB_PU_OSC_INJBO = '10.0.38.74'
    TS_PU_OSC_EJEBO = '10.0.38.77'
    TS_PU_OSC_INJSI = '10.0.38.20'
    SI_PU_OSC_INJSI = '10.0.38.69'


class ScopeSignals:
    """Mapping of physical signals to scope channels."""
    
    SI_FILL_PATTERN = (Scopes.AS_DI_FPMDIG, 5025, 'CHAN1')
    BO_FILL_PATTERN = (Scopes.AS_DI_FPMDIG, 5025, 'CHAN2')
    TS_EJESEPTG_PULSE = (Scopes.TS_PU_OSC_EJEBO, 5025, 'CHAN1')
    TS_EJESEPTF_PULSE = (Scopes.TS_PU_OSC_EJEBO, 5025, 'CHAN2')
    BO_EJEKCKR_PULSE = (Scopes.TS_PU_OSC_EJEBO, 5025, 'CHAN3')
    SI_PINGV_PULSE = (Scopes.TS_PU_OSC_EJEBO, 5025, 'CHAN4')

    @staticmethod
    def get_scope_name(scopesignal=None, scope_ip=None):
        """."""
        for symb in Scopes.__dict__:
            if symb[:2] in ('AS', 'LI', 'TB', 'BO', 'TS', 'SI'):
                scope_ip = getattr(Scopes, symb)
                if scope_ip == scopesignal[0] or scope_ip == scope_ip:
                    return symb







"""Define the PV database of a single BPM and its enum types."""
from copy import deepcopy as _dcopy
import numpy as _np

from ... import csdev as _csdev


# --- Enumeration Types ---
class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    TRIGDIR = ('trn', 'rcv')
    TRIGDIRPOL = ('same', 'rev')
    TRIGSRC = ('ext', 'int')
    TRIGEXTERN = (
        'UserDefined', 'BeamLossIn', 'BeamLossOut', 'OrbitInterlock',
        'StorageRingTbTClk', 'BoosterTbTClk', 'FOFBClk', 'Reserved')
    TRIGINTERN = ('FMCTrigIn', 'DSPSWClk')
    LOGTRIGINTERN = (
        'ADC', 'ADCSwap', 'MixerIQ', 'Unconnected', 'TbTIQ',
        'Unconnected2', 'TbTAmp', 'TbTPha', 'TbTPos', 'FOFBIQ',
        'Unconnected3', 'FOFBAmp', 'FOFBPha', 'FOFBPos', 'MonitAmp',
        'MonitPos', 'Monit1Pos', 'FMCTrigOut', 'Unconnected4', 'Unconnected5',
        'Unconnected6', 'Unconnected7', 'Unconnected8', 'Unconnected9')
    SWMODES = ('rffe_switching', 'direct', 'inverted', 'switching')
    SWTAGENBL = ('disabled', 'enabled')
    SWDATAMASKENBL = ('disabled', 'enabled')
    MONITENBL = ('No', 'Yes')

    OPMODES = ('MultiBunch', 'SinglePass')
    POLARITY = ('Positive', 'Negative')
    ENBLTYP = ('Disable', 'Enable')
    ENBLDDSBLD = ('disabled', 'enabled')
    CONNTYP = _csdev.ETypes.DISCONN_CONN
    ACQREPEAT = ('Normal', 'Repetitive')
    ACQEVENTS = ('Start', 'Stop', 'Abort')
    ACQDATATYP = ('A', 'B', 'C', 'D')
    ACQCHAN = ('ADC', 'ADCSwp', 'TbT', 'FOFB', 'TbTPha', 'FOFBPha', 'Monit1')
    ACQSTATES = (
        'Idle', 'Waiting', 'External Trig', 'Data Trig', 'Software Trig',
        'Acquiring', 'Error', 'Aborted', 'Too Many Samples',
        'Too Few Samples', 'No Memory', 'Acq Overflow')
    ACQTRIGTYP = ('Now', 'External', 'Data', 'Software')
    FFTWINDOWTYP = ('Square', 'Hanning', 'Parzen', 'Welch', 'QuadW')
    FFTCONVDIRECTION = ('Forward', 'Backward')
    FFTAVGSUBTRACT = ('No Subtraction', 'Average', 'Linear')
    FFTWRITABLEPROPS = ('INDX', 'MXIX', 'WIND', 'CDIR', 'ASUB', 'SPAN')


_et = ETypes  # syntactic sugar


# --- Const class ---
class Const(_csdev.Const):
    """Const class defining BPM constants."""

    TrigDir = _csdev.Const.register('TrigDir', _et.TRIGDIR)
    TrigDirPol = _csdev.Const.register('TrigDirPol', _et.TRIGDIRPOL)
    TrigSrc = _csdev.Const.register('TrigSrc', _et.TRIGSRC)
    TrigExtern = _csdev.Const.register('TrigExtern', _et.TRIGEXTERN)
    TrigIntern = _csdev.Const.register('TrigIntern', _et.TRIGINTERN)
    LogTrigIntern = _csdev.Const.register('LogTrigIntern', _et.LOGTRIGINTERN)
    SwModes = _csdev.Const.register('SwModes', _et.SWMODES)
    SwTagEnbl = _csdev.Const.register('SwTagEnbl', _et.SWTAGENBL)
    SwDataMaskEnbl = _csdev.Const.register(
        'SwDataMaskEnbl', _et.SWDATAMASKENBL)
    MonitEnbl = _csdev.Const.register('MonitEnbl', _et.MONITENBL)
    OpModes = _csdev.Const.register('OpModes', _et.OPMODES)
    Polarity = _csdev.Const.register('Polarity', _et.POLARITY)
    EnblTyp = _csdev.Const.register('EnblTyp', _et.ENBLTYP)
    EnbldDsbld = _csdev.Const.register('EnbldDsbld', _et.ENBLDDSBLD)
    ConnTyp = _csdev.Const.register('ConnTyp', _et.CONNTYP)
    AcqRepeat = _csdev.Const.register('AcqRepeat', _et.ACQREPEAT)
    AcqEvents = _csdev.Const.register('AcqEvents', _et.ACQEVENTS)
    AcqDataTyp = _csdev.Const.register('AcqDataTyp', _et.ACQDATATYP)
    AcqChan = _csdev.Const.register('AcqChan', _et.ACQCHAN)
    AcqStates = _csdev.Const.register('AcqStates', _et.ACQSTATES)
    AcqTrigTyp = _csdev.Const.register('AcqTrigTyp', _et.ACQTRIGTYP)
    FFTWindowTyp = _csdev.Const.register('FFTWindowTyp', _et.FFTWINDOWTYP)
    FFTConvDirection = _csdev.Const.register(
        'FFTConvDirection', _et.FFTCONVDIRECTION)
    FFTAvgSubtract = _csdev.Const.register(
        'FFTAvgSubtract', _et.FFTAVGSUBTRACT)
    FFTWritableProps = _csdev.Const.register(
        'FFTWritableProps', _et.FFTWRITABLEPROPS)

    @staticmethod
    def get_bpm_database(prefix=''):
        """Get the PV database of a single BPM."""
        data_db = {'type': 'float', 'value': 0.0, 'low': -1e12, 'high': 1e12}
        dbase = {
            'INFOFOFBRate-SP': {
                'type': 'int', 'value': 1910, 'low': 0, 'high': 2**31-1},
            'INFOFOFBRate-RB': {
                'type': 'int', 'value': 1910, 'low': 0, 'high': 2**31-1},
            'INFOHarmonicNumber-SP': {
                'type': 'int', 'value': 864, 'low': 0, 'high': 2**31-1},
            'INFOHarmonicNumber-RB': {
                'type': 'int', 'value': 864, 'low': 0, 'high': 2**31-1},
            'INFOMONITRate-SP': {
                'type': 'int', 'value': 21965000, 'low': 0, 'high': 2**31-1},
            'INFOMONITRate-RB': {
                'type': 'int', 'value': 21965000, 'low': 0, 'high': 2**31-1},
            'INFOMONIT1Rate-SP': {
                'type': 'int', 'value': 21965000, 'low': 0, 'high': 2**31-1},
            'INFOMONIT1Rate-RB': {
                'type': 'int', 'value': 21965000, 'low': 0, 'high': 2**31-1},
            'INFOTBTRate-SP': {
                'type': 'int', 'value': 382, 'low': 0, 'high': 2**31-1},
            'INFOTBTRate-RB': {
                'type': 'int', 'value': 382, 'low': 0, 'high': 2**31-1},
            }

        # PHYSICAL TRIGGERS
        for i in range(8):
            trig = 'TRIGGER{0:d}'.format(i)
            dbase.update(Const.get_physical_trigger_database(trig))

        # LOGICAL TRIGGERS
        for trig_tp in ('', '_PM'):
            for i in range(24):
                trig = 'TRIGGER' + trig_tp + '{0:d}'.format(i)
                dbase.update(Const.get_logical_trigger_database(trig))

        # AMPLITUDES AND POSITION CHANNELS
        for amp_tp in ('', 'SP'):
            dbase.update(Const.get_amplitudes_database(amp_tp))

        # SETTINGS AND STATUS
        dbase.update(Const.get_offsets_database())
        dbase.update(Const.get_gain_database())
        dbase.update(Const.get_rffe_database())
        dbase.update(Const.get_asyn_database())
        dbase.update(Const.get_switch_database())
        dbase.update(Const.get_monit_database())

        data_names = {
            'GEN': ['A', 'B', 'C', 'D', 'Q', 'SUM', 'X', 'Y'],
            'PM': ['A', 'B', 'C', 'D', 'Q', 'SUM', 'X', 'Y'],
            'SP': ['A', 'B', 'C', 'D'],
            }
        data_db = {
            'type': 'float', 'value': _np.array(100000*[0.0]), 'count': 100000}

        # ARRAY DATA FROM TRIGGERED ACQUISITIONS
        for acq_tp in ('GEN', 'SP', 'PM'):
            for prop in data_names[acq_tp]:
                nm = acq_tp + '_' + prop
                dbase[nm + 'ArrayData'] = _dcopy(data_db)
                dbase.update(Const.get_statistic_database(nm))
                if acq_tp == 'GEN':
                    dbase.update(Const.get_fft_database(nm))

        # TRIGGERED ACQUISITIONS CONFIGURATION
        for acq_md in ('ACQ', 'ACQ_PM'):
            dbase.update(Const.get_config_database(acq_md))

        for _, v in dbase.items():
            if 'low' in v:
                v['lolo'] = v['low']
                v['lolim'] = v['low']
            if 'high' in v:
                v['hihi'] = v['high']
                v['hilim'] = v['high']

        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_asyn_database(prefix=''):
        """."""
        dbase = {
            'asyn.ENBL': {
                'type': 'enum', 'enums': Const.EnblTyp._fields, 'value': 0},
            'asyn.CNCT': {
                'type': 'enum', 'enums': Const.ConnTyp._fields, 'value': 0},
            }
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_rffe_database(prefix=''):
        """."""
        prefix = prefix + 'RFFE'
        dbase = {
            'Att-SP': {
                'type': 'float', 'value': 0, 'unit': 'dB',
                'low': 0, 'high': 100},
            'Att-RB': {
                'type': 'float', 'value': 0, 'unit': 'dB',
                'low': 0, 'high': 100},
            'PidSpAC-SP': {
                'type': 'float', 'value': 0, 'unit': 'oC',
                'low': 0, 'high': 100},
            'PidSpAC-RB': {
                'type': 'float', 'value': 0, 'unit': 'oC',
                'low': 0, 'high': 100},
            'PidSpBD-SP': {
                'type': 'float', 'value': 0, 'unit': 'oC',
                'low': 0, 'high': 100},
            'PidSpBD-RB': {
                'type': 'float', 'value': 0, 'unit': 'oC',
                'low': 0, 'high': 100},
            'HeaterAC-SP': {
                'type': 'float', 'value': 0, 'unit': 'V',
                'low': 0, 'high': 100},
            'HeaterAC-RB': {
                'type': 'float', 'value': 0, 'unit': 'V',
                'low': 0, 'high': 100},
            'HeaterBD-SP': {
                'type': 'float', 'value': 0, 'unit': 'V',
                'low': 0, 'high': 100},
            'HeaterBD-RB': {
                'type': 'float', 'value': 0, 'unit': 'V',
                'low': 0, 'high': 100},
            'PidACKp-SP': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidACKp-RB': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidBDKp-SP': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidBDKp-RB': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidACTi-SP': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidACTi-RB': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidBDTi-SP': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidBDTi-RB': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidACTd-SP': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidACTd-RB': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidBDTd-SP': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            'PidBDTd-RB': {
                'type': 'float', 'value': 0, 'low': 0, 'high': 100},
            }
        dbase.update(Const.get_asyn_database())
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_switch_database(prefix=''):
        """."""
        dbase = {
            'SwMode-Sel': {
                'type': 'enum', 'enums': Const.SwModes._fields, 'value': 3},
            'SwMode-Sts': {
                'type': 'enum', 'enums': Const.SwModes._fields, 'value': 3},
            'SwTagEn-Sel': {
                'type': 'enum', 'enums': Const.SwTagEnbl._fields, 'value': 0},
            'SwTagEn-Sts': {
                'type': 'enum', 'enums': Const.SwTagEnbl._fields, 'value': 0},
            'SwDataMaskEn-Sel': {
                'type': 'enum', 'enums': Const.SwDataMaskEnbl._fields,
                'value': 0},
            'SwDataMaskEn-Sts': {
                'type': 'enum', 'enums': Const.SwDataMaskEnbl._fields,
                'value': 0},
            'SwDly-SP': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31-1},
            'SwDly-RB': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31-1},
            'SwDivClk-SP': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31-1},
            'SwDivClk-RB': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31-1},
            'SwDataMaskSamples-SP': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31-1},
            'SwDataMaskSamples-RB': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31-1},
            }
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_monit_database(prefix=''):
        """."""
        dbase = {
            'MonitEnable-Sel': {
                'type': 'enum', 'enums': Const.MonitEnbl._fields, 'value': 3},
            'MonitEnable-Sts': {
                'type': 'enum', 'enums': Const.MonitEnbl._fields, 'value': 3},
            'MONITUpdtTime-SP': {
                'type': 'float', 'value': 0, 'low': 0.05, 'high': 1.0,
                'unit': 's'},
            'MONITUpdtTime-RB': {
                'type': 'float', 'value': 0, 'low': 0.05, 'high': 1.0,
                'unit': 's'},
            }
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_offsets_database(prefix=''):
        """."""
        data_db = {'type': 'float', 'value': 0.0, 'low': -1e12, 'high': 1e12}
        dbase = {
            'PosQOffset-SP': _dcopy(data_db), 'PosQOffset-RB': _dcopy(data_db),
            'PosXOffset-SP': _dcopy(data_db), 'PosXOffset-RB': _dcopy(data_db),
            'PosYOffset-SP': _dcopy(data_db), 'PosYOffset-RB': _dcopy(data_db),
            }
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_gain_database(prefix=''):
        """."""
        data_db = {'type': 'float', 'value': 0.0, 'low': -1e12, 'high': 1e12}
        dbase = {
            'PosKq-SP': _dcopy(data_db), 'PosKq-RB': _dcopy(data_db),
            'PosKsum-SP': _dcopy(data_db), 'PosKsum-RB': _dcopy(data_db),
            'PosKx-SP': _dcopy(data_db), 'PosKx-RB': _dcopy(data_db),
            'PosKy-SP': _dcopy(data_db), 'PosKy-RB': _dcopy(data_db),
            }
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_amplitudes_database(prefix=''):
        """."""
        data_db = {'type': 'float', 'value': 0.0, 'low': -1e12, 'high': 1e12}
        dbase = {
            'PosX-Mon': _dcopy(data_db), 'PosY-Mon': _dcopy(data_db),
            'Sum-Mon': _dcopy(data_db), 'PosQ-Mon': _dcopy(data_db),
            'AmplA-Mon': _dcopy(data_db), 'AmplB-Mon': _dcopy(data_db),
            'AmplC-Mon': _dcopy(data_db), 'AmplD-Mon': _dcopy(data_db),
            }
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_physical_trigger_database(prefix=''):
        """."""
        dbase = {
            'Dir-Sel': {
                'type': 'enum', 'enums': Const.TrigDir._fields, 'value': 1},
            'Dir-Sts': {
                'type': 'enum', 'enums': Const.TrigDir._fields, 'value': 1},
            'DirPol-Sel': {
                'type': 'enum', 'enums': Const.TrigDirPol._fields, 'value': 1},
            'DirPol-Sts': {
                'type': 'enum', 'enums': Const.TrigDirPol._fields, 'value': 1},
            'RcvCntRst-SP': {
                'type': 'int', 'value': 0},
            'TrnCntRst-SP': {
                'type': 'int', 'value': 0},
            'RcvCnt-Mon': {
                'type': 'int', 'value': 0},
            'TrnCnt-Mon': {
                'type': 'int', 'value': 0},
            'RcvLen-SP': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**15-1},
            'RcvLen-RB': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**15-1},
            'TrnLen-SP': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**15-1},
            'TrnLen-RB': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**15-1},
            }
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_logical_trigger_database(prefix=''):
        """."""
        dbase = {
            'RcvSrc-Sel': {
                'type': 'enum', 'enums': Const.TrigSrc._fields, 'value': 0},
            'RcvSrc-Sts': {
                'type': 'enum', 'enums': Const.TrigSrc._fields, 'value': 0},
            'TrnSrc-Sel': {
                'type': 'enum', 'enums': Const.TrigSrc._fields, 'value': 0},
            'TrnSrc-Sts': {
                'type': 'enum', 'enums': Const.TrigSrc._fields, 'value': 0},
            'RcvInSel-SP': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**15-1},
            'RcvInSel-RB': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**15-1},
            'TrnOutSel-SP': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**15-1},
            'TrnOutSel-RB': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**15-1},
            }
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_config_database(prefix=''):
        """Get the configuration PVs database."""
        dbase = {
            'BPMMode-Sel': {
                'type': 'enum', 'enums': Const.OpModes._fields, 'value': 0},
            'BPMMode-Sts': {
                'type': 'enum', 'enums': Const.OpModes._fields, 'value': 0},
            'Channel-Sel': {
                'type': 'enum', 'enums': Const.AcqChan._fields, 'value': 0},
            'Channel-Sts': {
                'type': 'enum', 'enums': Const.AcqChan._fields, 'value': 0},
            # 'NrShots-SP': {
            'Shots-SP': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 65536},
            # 'NrShots-RB': {
            'Shots-RB': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 65536},
            'TriggerHwDly-SP': {
                'type': 'float', 'value': 0.0, 'low': 0.0, 'high': 1e9},
            'TriggerHwDly-RB': {
                'type': 'float', 'value': 0.0, 'low': 0.0, 'high': 1e9},
            'UpdateTime-SP': {
                'type': 'float', 'value': 1.0, 'low': 0.001, 'high': 1e9},
            'UpdateTime-RB': {
                'type': 'float', 'value': 1.0, 'low': 0.001, 'high': 1e9},
            # 'NrSamplesPre-SP': {
            'SamplesPre-SP': {
                'type': 'int', 'value': 1000, 'low': 0, 'high': 100000},
            # 'NrSamplesPre-RB': {
            'SamplesPre-RB': {
                'type': 'int', 'value': 1000, 'low': 0, 'high': 100000},
            # 'NrSamplesPost-SP': {
            'SamplesPost-SP': {
                'type': 'int', 'value': 1000, 'low': 0, 'high': 100000},
            # 'NrSamplesPost-RB': {
            'SamplesPost-RB': {
                'type': 'int', 'value': 1000, 'low': 0, 'high': 100000},
            # 'Ctrl-Sel': {
            'TriggerEvent-Sel': {
                'type': 'enum', 'enums': Const.AcqEvents._fields,
                'value': Const.AcqEvents.Stop},
            # 'Ctrl-Sts': {
            'TriggerEvent-Sts': {
                'type': 'enum', 'enums': Const.AcqEvents._fields,
                'value': Const.AcqEvents.Stop},
            # 'Status-Mon': {
            'Status-Sel': {
                'type': 'enum', 'enums': Const.AcqStates._fields, 'value': 0},
            # 'TriggerType-Sel': {
            'Trigger-Sel': {
                'type': 'enum', 'enums': Const.AcqTrigTyp._fields, 'value': 1},
            # 'TriggerType-Sts': {
            'Trigger-Sts': {
                'type': 'enum', 'enums': Const.AcqTrigTyp._fields, 'value': 1},
            'TriggerRep-Sel': {
                'type': 'enum', 'enums': Const.AcqRepeat._fields, 'value': 0},
            'TriggerRep-Sts': {
                'type': 'enum', 'enums': Const.AcqRepeat._fields, 'value': 0},
            # 'TriggerDataChan-Sel': {
            'DataTrigChan-Sel': {
                'type': 'enum', 'enums': Const.AcqChan._fields, 'value': 0},
            # 'TriggerDataChan-Sts': {
            'DataTrigChan-Sts': {
                'type': 'enum', 'enums': Const.AcqChan._fields, 'value': 0},
            # 'TriggerDataSel-Sel': {
            #     'type': 'enum', 'enums': AcqDataTyp._fields, 'value': 0},
            # 'TriggerDataSel-Sts': {
            #     'type': 'enum', 'enums': AcqDataTyp._fields, 'value': 0},
            'TriggerDataSel-SP': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 4},
            'TriggerDataSel-RB': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 4},
            'TriggerDataThres-SP': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**31 - 1},
            'TriggerDataThres-RB': {
                'type': 'int', 'value': 1, 'low': 0, 'high': 2**31 - 1},
            'TriggerDataPol-Sel': {
                'type': 'enum', 'enums': Const.Polarity._fields, 'value': 0},
            'TriggerDataPol-Sts': {
                'type': 'enum', 'enums': Const.Polarity._fields, 'value': 0},
            'TriggerDataHyst-SP': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31 - 1},
            'TriggerDataHyst-RB': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31 - 1},
            'TbtTagEn-Sel': {
                'type': 'enum', 'enums': Const.EnbldDsbld._fields, 'value': 0},
            'TbtTagEn-Sts': {
                'type': 'enum', 'enums': Const.EnbldDsbld._fields, 'value': 0},
            'TbtTagDly-SP': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31 - 1},
            'TbtTagDly-RB': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31 - 1},
            'TbtDataMaskEn-Sel': {
                'type': 'enum', 'enums': Const.EnbldDsbld._fields, 'value': 0},
            'TbtDataMaskEn-Sts': {
                'type': 'enum', 'enums': Const.EnbldDsbld._fields, 'value': 0},
            'TbtDataMaskSamplesBeg-SP': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31 - 1},
            'TbtDataMaskSamplesBeg-RB': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31 - 1},
            'TbtDataMaskSamplesEnd-SP': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31 - 1},
            'TbtDataMaskSamplesEnd-RB': {
                'type': 'int', 'value': 0, 'low': 0, 'high': 2**31 - 1},
            }
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_fft_database(prefix=''):
        """Get the PV database of the FFT plugin."""
        data_db = {
            'type': 'float', 'value': _np.array(100000*[0.0]), 'count': 100000}
        acq_int_db = {'type': 'int', 'value': 1, 'low': 0, 'high': 100000}
        dbase = dict()
        dbase['FFTFreq-Mon'] = _dcopy(data_db)
        dbase['FFTData.SPAN'] = _dcopy(acq_int_db)
        dbase['FFTData.AMP'] = _dcopy(data_db)
        dbase['FFTData.PHA'] = _dcopy(data_db)
        dbase['FFTData.SIN'] = _dcopy(data_db)
        dbase['FFTData.COS'] = _dcopy(data_db)
        dbase['FFTData.WAVN'] = _dcopy(data_db)
        dbase['FFTData.INDX'] = _dcopy(acq_int_db)
        dbase['FFTData.MXIX'] = _dcopy(acq_int_db)
        dbase['FFTData.WIND'] = {
            'type': 'enum', 'enums': Const.FFTWindowTyp._fields, 'value': 0}
        dbase['FFTData.CDIR'] = {
            'type': 'enum', 'enums': Const.FFTConvDirection._fields,
            'value': 0}
        dbase['FFTData.ASUB'] = {
            'type': 'enum', 'enums': Const.FFTAvgSubtract._fields, 'value': 0}
        return {prefix + k: v for k, v in dbase.items()}

    @staticmethod
    def get_statistic_database(prefix=''):
        """Get the PV database of the STAT plugin."""
        acq_data_stat_db = {
            'type': 'float', 'value': 0.0, 'low': -1e12, 'high': 1e12}
        dbase = dict()
        dbase['_STATSMaxValue_RBV'] = _dcopy(acq_data_stat_db)
        dbase['_STATSMeanValue_RBV'] = _dcopy(acq_data_stat_db)
        dbase['_STATSMinValue_RBV'] = _dcopy(acq_data_stat_db)
        dbase['_STATSSigma_RBV'] = _dcopy(acq_data_stat_db)
        return {prefix + k: v for k, v in dbase.items()}



"""."""

from functools import partial as _part
from epics import PV as _PV

from ...thread import RepeaterThread as _Repeater
from ...clientconfigdb import ConfigDBClient as _ConfigDBClient
from ..util import BaseClass as _BaseClass, ProcessImage as _ProcessImage
from .csdev import Const as _Const


class MeasParameters(_BaseClass, _Const):
    """."""

    def __init__(self, callback=None):
        """."""
        self.image_processor = _ProcessImage(callback=callback)
        self.image_processor.px2mmscalex = self.DEF_COEFX
        self.image_processor.px2mmscaley = self.DEF_COEFY
        self.image_processor.roiautocenter = self.AutoCenter.Manual
        self.image_processor.imageflipx = self.image_processor.ImgFlip.On
        self.image_processor.imageflipy = self.image_processor.ImgFlip.Off
        self.image_processor.roisizex = self.DEF_ROISIZE
        self.image_processor.roisizey = self.DEF_ROISIZE
        self.image_processor.roicenterx = self.TARGETX
        self.image_processor.roicentery = self.TARGETY
        self.image_processor.method = self.Method.Moments
        self._profile = self.DEF_PROFILE
        self._target_posx = self.TARGETX
        self._target_posy = self.TARGETY
        self._needed_dbumpx = 0.0
        self._needed_dbumpy = 0.0
        self._applied_bumpx = 0.0
        self._applied_bumpy = 0.0
        self._need_update = True
        self._apply_status = self.ApplyStatus.Idle
        conf = _ConfigDBClient(config_type='si_orbit')
        self._refs = conf.get_config_value('ref_orb')
        self._sofb_pvs = dict(
            refx_sp=_PV('SI-Glob:AP-SOFB:RefOrbX-SP'),
            refy_sp=_PV('SI-Glob:AP-SOFB:RefOrbY-SP'),
            refx_rb=_PV(
                'SI-Glob:AP-SOFB:RefOrbX-RB', callback=self._set_need_update),
            refy_rb=_PV(
                'SI-Glob:AP-SOFB:RefOrbY-RB', callback=self._set_need_update))

        self._width_source = _PV(self._profile + 'cam1:ArraySizeX_RBV')
        self._image_source = _PV(
            self._profile+'image1:ArrayData', auto_monitor=True)
        super().__init__(callback=callback)
        self._thread = _Repeater(1/self.DEF_RATE, self.process_image, niter=0)
        self._thread.pause()
        self._thread.start()

    def get_map2write(self):
        """."""
        dic_ = self.image_processor.get_map2write()
        dic_.update({
            'ApplyBump-Cmd': self.cmd_apply_bump,
            'MeasureCtrl-Sel': _part(self.write, 'measuring'),
            'MeasureRate-SP': _part(self.write, 'rate'),
            'TargetPosX-SP': _part(self.write, 'target_posx'),
            'TargetPosY-SP': _part(self.write, 'target_posy'),
            })
        return dic_

    def get_map2read(self):
        """."""
        dic_ = self.image_processor.get_map2read()
        dic_.update({
            'MeasureCtrl-Sts': _part(self.read, 'measuring'),
            'MeasureRate-RB': _part(self.read, 'rate'),
            'TargetPosX-RB': _part(self.read, 'target_posx'),
            'TargetPosY-RB': _part(self.read, 'target_posy'),
            'NeededDeltaBumpX-Mon': _part(self.read, 'needed_dbumpx'),
            'NeededDeltaBumpY-Mon': _part(self.read, 'needed_dbumpy'),
            'AppliedBumpX-Mon': _part(self.read, 'applied_bumpx'),
            'AppliedBumpY-Mon': _part(self.read, 'applied_bumpy'),
            'ApplyStatus-Mon': _part(self.read, 'apply_status'),
            })
        return dic_

    def start(self):
        """."""
        self._thread.resume()
        self.run_callbacks('MeasureCtrl-Sts', 1)

    def stop(self):
        """."""
        self._thread.pause()
        self.run_callbacks('MeasureCtrl-Sts', 0)

    @property
    def connected(self):
        """."""
        conn = self._width_source.connected
        conn &= self._image_source.connected
        conn &= all(map(lambda x: x.connected, self._sofb_pvs.values()))
        return conn

    @property
    def rate(self):
        """."""
        return 1/self._thread.interval

    @rate.setter
    def rate(self, val):
        """."""
        if isinstance(val, (float, int)) and 0 < val < 30:
            self._thread.interval = 1/val
            self.run_callbacks('MeasureRate-RB', val)

    @property
    def apply_status(self):
        """Return status of last apply attempt."""
        return self._apply_status

    @property
    def measuring(self):
        """."""
        return not self._thread.is_paused()

    @measuring.setter
    def measuring(self, val):
        """."""
        if val:
            self.start()
        else:
            self.stop()

    @property
    def target_posx(self):
        """."""
        return self._target_posx

    @target_posx.setter
    def target_posx(self, val):
        """."""
        if isinstance(val, (float, int)):
            self._target_posx = val
            self.run_callbacks('TargetPosX-RB', val)

    @property
    def target_posy(self):
        """."""
        return self._target_posy

    @target_posy.setter
    def target_posy(self, val):
        """."""
        if isinstance(val, (float, int)):
            self._target_posy = val
            self.run_callbacks('TargetPosY-RB', val)

    @property
    def needed_dbumpx(self):
        """."""
        return self._needed_dbumpx

    @property
    def needed_dbumpy(self):
        """."""
        return self._needed_dbumpy

    @property
    def applied_bumpx(self):
        """."""
        return self._applied_bumpx

    @property
    def applied_bumpy(self):
        """."""
        return self._applied_bumpy

    def cmd_apply_bump(self, *args):
        """Apply Bump to SOFB."""
        _ = args
        if not self.connected:
            self._apply_status = self.ApplyStatus.ConnectionError
            self.run_callbacks('ApplyStatus-Mon', self._apply_status)
            return
        self._update_applied_bumps()
        self._apply_status = self._calc_and_apply_bumps()
        self.run_callbacks('ApplyStatus-Mon', self._apply_status)

    def process_image(self):
        """."""
        if self._need_update:
            self._update_applied_bumps()

        value = self._width_source.value
        if isinstance(value, (float, int)):
            self.image_processor.imagewidth = int(value)

        self.image_processor.image = self._image_source.value
        dltx = self.image_processor.beamcenterx - self._target_posx
        dlty = self.image_processor.beamcentery - self._target_posy
        dltx *= self.image_processor.px2mmscalex * 1e-3  # mm --> m
        dlty *= self.image_processor.px2mmscaley * 1e-3  # mm --> m

        self._needed_dbumpx = -dltx / self.DIST_FROM_SRC * 1e6  # rad --> urad
        # NOTE: this signal asymmetry is due to the x-flip of the image and
        # the beamline optics.
        self._needed_dbumpy = dlty / self.DIST_FROM_SRC * 1e6  # rad --> urad

        self.run_callbacks('NeededDeltaBumpX-Mon', self._needed_dbumpx)
        self.run_callbacks('NeededDeltaBumpY-Mon', self._needed_dbumpy)

    def _update_applied_bumps(self):
        if not self.connected:
            return
        orbx = self._sofb_pvs['refx_rb'].value
        orby = self._sofb_pvs['refy_rb'].value
        if orbx is None or orby is None:
            return

        dx1 = orbx[self.BPM1_INDEX] - self._refs['x'][self.BPM1_INDEX]
        dx2 = orbx[self.BPM2_INDEX] - self._refs['x'][self.BPM2_INDEX]
        dy1 = orby[self.BPM1_INDEX] - self._refs['y'][self.BPM1_INDEX]
        dy2 = orby[self.BPM2_INDEX] - self._refs['y'][self.BPM2_INDEX]

        self._applied_bumpx = (dx2 - dx1) / self.SS_SIZE  # already in [urad]
        self._applied_bumpy = (dy2 - dy1) / self.SS_SIZE

        self.run_callbacks('AppliedBumpX-Mon', self._applied_bumpx)
        self.run_callbacks('AppliedBumpY-Mon', self._applied_bumpy)
        self._need_update = False

    def _calc_and_apply_bumps(self):
        dbumpx = self._needed_dbumpx
        dbumpy = self._needed_dbumpy
        total_bumpx = dbumpx + self._applied_bumpx
        total_bumpy = dbumpy + self._applied_bumpy

        maxb = max(abs(total_bumpx), abs(total_bumpy))
        if maxb > self.MAX_BUMP:
            return self.ApplyStatus.LimitExceeded

        dax = dbumpx * self.SS_SIZE / 2
        day = dbumpy * self.SS_SIZE / 2

        orbx = self._sofb_pvs['refx_rb'].value.copy()
        orby = self._sofb_pvs['refy_rb'].value.copy()
        orbx[self.BPM1_INDEX] -= dax
        orby[self.BPM1_INDEX] -= day
        orbx[self.BPM2_INDEX] += dax
        orby[self.BPM2_INDEX] += day

        self._sofb_pvs['refx_sp'].value = orbx
        self._sofb_pvs['refy_sp'].value = orby
        return self.ApplyStatus.Applied

    def _set_need_update(self, *args, **kwargs):
        _ = args
        _ = kwargs
        self._need_update = True



"""SI tune correction configuration.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict)
    }
    return _dict


#  Tune Correction Parameters for Storage Ring
#
#  | DeltaTuneX |    | m00  m01...m07 |   | KL SI QFA  |
#  |            | =  |                | * |      .     |
#  | DeltaTuneY |    | m10  m11...m17 |   |      .     |
#                                         |      .     |
#                                         | KL SI QDP2 |
# Where (1+f)KL = KL + DeltaKL.
#
# Correction Matrix of Svd and Additional Method
# (obtained by matlab lnls_calc_tunecorr_params routine)
#   m(0,0)   m(0,1)...m(0,7)
#   m(1,0)   m(1,1)...m(1,7)
#
# Nominals KLs
# [quadrupole_order   QFA  QFB  QFP  QDA  QDB1  QDB2  QDP1  QDP2]

_template_dict = {
    'matrix': [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
               [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]],
    'nominal KLs': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
}



#!/usr/bin/env python-sirius

"""Test the configuration client class."""
from unittest import TestCase

from siriuspy.clientconfigdb import ConfigDBClient, ConfigDBDocument

import siriuspy.util as util


class TestConfigDBClient(TestCase):
    """Test update and delete config meets requirements."""

    api = {
        'config_type',
        "url",
        "connected",
        "get_dbsize",
        "get_nrconfigs",
        "get_config_types",
        "get_config_types_from_templates",
        "find_configs",
        "get_config_value",
        "get_config_info",
        "rename_config",
        "insert_config",
        "delete_config",
        "retrieve_config",
        "get_value_from_template",
        'check_valid_value',
        'check_valid_configname',
        "conv_timestamp_txt_2_flt",
        "conv_timestamp_flt_2_txt",
    }

    def test_api(self):
        """Test api."""
        valid = util.check_public_interface_namespace(
            ConfigDBClient, TestConfigDBClient.api)
        self.assertTrue(valid)

class TestConfigDBDocument(TestCase):
    """Test update and delete config meets requirements."""

    api = {
        'configdbclient',
        'url',
        'config_type',
        'connected',
        'name',
        'name',
        'info',
        'created',
        'modified',
        'modified_count',
        'discarded',
        'value',
        'value',
        'synchronized',
        'exist',
        'load',
        'save',
        'delete',
        'check_valid_value',
        'get_value_from_template',
        'generate_config_name',
    }

    def test_api(self):
        """Test api."""
        valid = util.check_public_interface_namespace(
            ConfigDBDocument, TestConfigDBDocument.api)
        self.assertTrue(valid)


class TestConfigServiceConTimestamp(TestCase):
    """Test response error handling."""

    def _test_conv_timestamp(self):
        """Test timestamp conversion."""
        # TODO: NOT WORKING ON TRAVIS
        samples = {
            ("Dec 11, 2017", 1512957600.0),
            ("12/11/2017", 1512957600.0),
            ("2017/12/11", 1512957600.0),
            ("2017-12-11", 1512957600.0),
            ("Dec 11 2017 14:00:00", 1513008000.0),
            ("12/11/2017 14:00:00", 1513008000.0),
            ("2017/12/11 14:00:00", 1513008000.0),
            ("2017-12-11 14:00:00", 1513008000.0),
            ("2017-12-11T14:00:00", 1513008000.0),
            ("2017-12-11 14:00:00+01:00", 1512997200.0),
            ("2017-12-11T14:00:00+01:00", 1512997200.0),
            ("2017-12-11T14:00:00.45", 1513008000.45),
        }

        for sample in samples:
            date_string = sample[0]
            timestamp = sample[1]
            self.assertEqual(
                ConfigDBClient.conv_timestamp_txt_2_flt(date_string),
                timestamp)



"""."""

from functools import partial as _part
import logging as _log
import numpy as _np
from scipy.optimize import curve_fit

from ..callbacks import Callback

from .csdev import Const as _Const


class BaseClass(Callback, _Const):
    """."""

    def __init__(self, callback=None):
        """."""
        self._map2read = self.get_map2read()
        self._map2write = self.get_map2write()
        super().__init__(callback=callback)

    def get_map2write(self):
        """."""
        return dict()

    def get_map2read(self):
        """."""
        return dict()

    def get(self, prop):
        """."""
        if prop in self._map2read:
            return self._map2read[prop]()
        return self.__getattribute__(prop)

    def read(self, prop):
        """."""
        return self.get(prop)

    def write(self, prop, value):
        """."""
        if prop in self._map2write:
            return self._map2write[prop](value)
        try:
            self.__setattr__(prop, value)
        except Exception:
            return False
        return True


class ProcessImage(BaseClass):
    """."""

    def __init__(self, callback=None):
        """."""
        super().__init__(callback=callback)
        self._roi_autocenter = True
        self._roi_cen = [0, 0]
        self._roi_size = [self.DEFAULT_ROI_SIZE, self.DEFAULT_ROI_SIZE]
        self._roi_start = [0, 0]
        self._roi_end = [0, 0]
        self._roi_axis = [
            _np.arange(self.DEFAULT_ROI_SIZE, dtype=int),
            _np.arange(self.DEFAULT_ROI_SIZE, dtype=int)]
        self._roi_proj = [
            _np.zeros(self.DEFAULT_ROI_SIZE, dtype=float),
            _np.zeros(self.DEFAULT_ROI_SIZE, dtype=float)]
        self._roi_gauss = [
            _np.zeros(self.DEFAULT_ROI_SIZE, dtype=float),
            _np.zeros(self.DEFAULT_ROI_SIZE, dtype=float)]
        self._background = _np.zeros(
            (self.DEFAULT_WIDTH, self.DEFAULT_HEIGHT), dtype=float)
        self._background_use = False
        self._crop = [0, 2**16]
        self._crop_use = False
        self._width = 0
        self._reading_order = self.ReadingOrder.CLike
        self._method = self.Method.GaussFit
        self._nr_averages = 1
        self._images = []
        self._reset_buffer = False
        self._image = _np.zeros(
            (self.DEFAULT_WIDTH, self.DEFAULT_HEIGHT), dtype=float)
        self._conv_autocenter = True
        self._conv_cen = [0, 0]
        self._conv_scale = [1, 1]
        self._flip = [False, False]
        self._beam_params = [None, None]

    def get_map2write(self):
        """."""
        return {
            'ResetBuffer-Cmd': self.cmd_reset_buffer,
            'Image-SP': _part(self.write, 'image'),
            'Width-SP': _part(self.write, 'imagewidth'),
            'ReadingOrder-Sel': _part(self.write, 'readingorder'),
            'NrAverages-SP': _part(self.write, 'nr_averages'),
            'ImgCropLow-SP': _part(self.write, 'imagecroplow'),
            'ImgCropHigh-SP': _part(self.write, 'imagecrophigh'),
            'ImgCropUse-Sel': _part(self.write, 'useimagecrop'),
            'ImgFlipX-Sel': _part(self.write, 'imageflipx'),
            'ImgFlipY-Sel': _part(self.write, 'imageflipy'),
            'CalcMethod-Sel': _part(self.write, 'method'),
            'ROIAutoCenter-Sel': _part(self.write, 'roiautocenter'),
            'ROICenterX-SP': _part(self.write, 'roicenterx'),
            'ROICenterY-SP': _part(self.write, 'roicentery'),
            'ROISizeX-SP': _part(self.write, 'roisizex'),
            'ROISizeY-SP': _part(self.write, 'roisizey'),
            'Background-SP': _part(self.write, 'background'),
            'BgUse-Sel': _part(self.write, 'usebackground'),
            'Px2mmScaleX-SP': _part(self.write, 'px2mmscalex'),
            'Px2mmScaleY-SP': _part(self.write, 'px2mmscaley'),
            'Px2mmAutoCenter-Sel': _part(self.write, 'px2mmautocenter'),
            'Px2mmCenterX-SP': _part(self.write, 'px2mmcenterx'),
            'Px2mmCenterY-SP': _part(self.write, 'px2mmcentery'),
            }

    def get_map2read(self):
        """."""
        return {
            'Image-RB': _part(self.read, 'image'),
            'Width-RB': _part(self.read, 'imagewidth'),
            'ReadingOrder-Sts': _part(self.read, 'readingorder'),
            'NrAverages-RB': _part(self.read, 'nr_averages'),
            'BufferSize-Mon': _part(self.read, 'buffer_size'),
            'ImgCropLow-RB': _part(self.read, 'imagecroplow'),
            'ImgCropHigh-RB': _part(self.read, 'imagecrophigh'),
            'ImgCropUse-Sts': _part(self.read, 'useimagecrop'),
            'ImgFlipX-Sts': _part(self.read, 'imageflipx'),
            'ImgFlipY-Sts': _part(self.read, 'imageflipy'),
            'CalcMethod-Sts': _part(self.read, 'method'),
            'ROIAutoCenter-Sts': _part(self.read, 'roiautocenter'),
            'ROICenterX-RB': _part(self.read, 'roicenterx'),
            'ROICenterY-RB': _part(self.read, 'roicentery'),
            'ROISizeX-RB': _part(self.read, 'roisizex'),
            'ROISizeY-RB': _part(self.read, 'roisizey'),
            'ROIStartX-Mon': _part(self.read, 'roistartx'),
            'ROIStartY-Mon': _part(self.read, 'roistarty'),
            'ROIEndX-Mon': _part(self.read, 'roiendx'),
            'ROIEndY-Mon': _part(self.read, 'roiendy'),
            'ROIProjX-Mon': _part(self.read, 'roiprojx'),
            'ROIProjY-Mon': _part(self.read, 'roiprojy'),
            'ROIAxisX-Mon': _part(self.read, 'roiaxisx'),
            'ROIAxisY-Mon': _part(self.read, 'roiaxisy'),
            'ROIGaussFitX-Mon': _part(self.read, 'roigaussx'),
            'ROIGaussFitY-Mon': _part(self.read, 'roigaussy'),
            'BeamCenterX-Mon': _part(self.read, 'beamcenterx'),
            'BeamCenterY-Mon': _part(self.read, 'beamcentery'),
            'BeamSizeX-Mon': _part(self.read, 'beamsizex'),
            'BeamSizeY-Mon': _part(self.read, 'beamsizey'),
            'BeamCentermmX-Mon': _part(self.read, 'beamcentermmx'),
            'BeamCentermmY-Mon': _part(self.read, 'beamcentermmy'),
            'BeamSizemmX-Mon': _part(self.read, 'beamsizemmx'),
            'BeamSizemmY-Mon': _part(self.read, 'beamsizemmy'),
            'BeamAmplX-Mon': _part(self.read, 'beamamplx'),
            'BeamAmplY-Mon': _part(self.read, 'beamamply'),
            'BeamOffsetX-Mon': _part(self.read, 'beamoffsetx'),
            'BeamOffsetY-Mon': _part(self.read, 'beamoffsety'),
            'Background-RB': _part(self.read, 'background'),
            'BgUse-Sts': _part(self.read, 'usebackground'),
            'Px2mmScaleX-RB': _part(self.read, 'px2mmscalex'),
            'Px2mmScaleY-RB': _part(self.read, 'px2mmscaley'),
            'Px2mmAutoCenter-Sts': _part(self.read, 'px2mmautocenter'),
            'Px2mmCenterX-RB': _part(self.read, 'px2mmcenterx'),
            'Px2mmCenterY-RB': _part(self.read, 'px2mmcentery'),
            }

    @property
    def image(self):
        """."""
        return self._image.copy().ravel()

    @image.setter
    def image(self, val):
        """."""
        if not isinstance(val, _np.ndarray):
            _log.error('Image is not a numpy array')
            return
        self._process_image(val)

    @property
    def imagewidth(self):
        """."""
        return self._width

    @imagewidth.setter
    def imagewidth(self, val):
        """."""
        if val is None:
            _log.error('could not set width')
            return
        self._width = int(val)
        self.run_callbacks('Width-RB', int(val))
        img = self._adjust_image_dimensions(self._background)
        if img is not None:
            self._background = img
        else:
            _log.error('could not set background')

    @property
    def readingorder(self):
        """."""
        return self._reading_order

    @readingorder.setter
    def readingorder(self, val):
        """."""
        if int(val) in self.ReadingOrder:
            self._reading_order = int(val)
            self.run_callbacks('ReadingOrder-Sts', int(val))

    @property
    def nr_averages(self):
        """Return the number of averages to be used."""
        return self._nr_averages

    @nr_averages.setter
    def nr_averages(self, val):
        self._nr_averages = int(val)
        self.run_callbacks('NrAverages-RB', val)

    @property
    def buffer_size(self):
        """Return the current buffer size."""
        return len(self._images)

    @property
    def imagecroplow(self):
        """."""
        return self._crop[self.CropIdx.Low]

    @imagecroplow.setter
    def imagecroplow(self, val):
        """."""
        val = int(val)
        if 0 <= val < self._crop[self.CropIdx.High]:
            self._crop[self.CropIdx.Low] = val
            self.run_callbacks('ImgCropLow-RB', val)

    @property
    def imagecrophigh(self):
        """."""
        return self._crop[self.CropIdx.High]

    @imagecrophigh.setter
    def imagecrophigh(self, val):
        """."""
        val = int(val)
        if self._crop[self.CropIdx.Low] < val:
            self._crop[self.CropIdx.High] = val
            self.run_callbacks('ImgCropHigh-RB', val)

    @property
    def useimagecrop(self):
        """."""
        return self._crop_use

    @useimagecrop.setter
    def useimagecrop(self, val):
        """."""
        self._crop_use = bool(val)
        self.run_callbacks('ImgCropUse-Sts', int(val))

    @property
    def imageflipx(self):
        """."""
        return self._flip[self.Plane.X]

    @imageflipx.setter
    def imageflipx(self, val):
        """."""
        self._flip[self.Plane.X] = bool(val)
        self.run_callbacks('ImgFlipX-Sts', int(val))

    @property
    def imageflipy(self):
        """."""
        return self._flip[self.Plane.Y]

    @imageflipy.setter
    def imageflipy(self, val):
        """."""
        self._flip[self.Plane.Y] = bool(val)
        self.run_callbacks('ImgFlipY-Sts', int(val))

    @property
    def imagesizex(self):
        """."""
        return self._image.shape[self.Plane.X]

    @property
    def imagesizey(self):
        """."""
        return self._image.shape[self.Plane.Y]

    @property
    def method(self):
        """."""
        return self._method

    @method.setter
    def method(self, val):
        """."""
        if int(val) in self.Method:
            self._method = int(val)
            self.run_callbacks('CalcMethod-Sts', int(val))

    @property
    def roiautocenter(self):
        """."""
        return self._roi_autocenter

    @roiautocenter.setter
    def roiautocenter(self, val):
        """."""
        self._roi_autocenter = bool(val)
        self.run_callbacks('ROIAutoCenter-Sts', int(val))

    @property
    def roicenterx(self):
        """."""
        return self._roi_cen[self.Plane.X]

    @roicenterx.setter
    def roicenterx(self, val):
        """."""
        val = int(val)
        if 0 <= val < self._image.shape[self.Plane.X]:
            self._roi_cen[self.Plane.X] = val
            self.run_callbacks('ROICenterX-RB', val)

    @property
    def roicentery(self):
        """."""
        return self._roi_cen[self.Plane.Y]

    @roicentery.setter
    def roicentery(self, val):
        """."""
        val = int(val)
        if 0 <= val < self._image.shape[self.Plane.Y]:
            self._roi_cen[self.Plane.Y] = val
            self.run_callbacks('ROICenterY-RB', val)

    @property
    def roisizex(self):
        """."""
        return self._roi_size[self.Plane.X]

    @roisizex.setter
    def roisizex(self, val):
        """."""
        val = int(val)
        if 1 <= val < self._image.shape[self.Plane.X]:
            self._roi_size[self.Plane.X] = val
            self.run_callbacks('ROISizeX-RB', val)

    @property
    def roisizey(self):
        """."""
        return self._roi_size[self.Plane.Y]

    @roisizey.setter
    def roisizey(self, val):
        """."""
        val = int(val)
        if 1 <= val < self._image.shape[self.Plane.Y]:
            self._roi_size[self.Plane.Y] = val
            self.run_callbacks('ROISizeY-RB', val)

    @property
    def roistartx(self):
        """."""
        return self._roi_start[self.Plane.X]

    @property
    def roistarty(self):
        """."""
        return self._roi_start[self.Plane.Y]

    @property
    def roiendx(self):
        """."""
        return self._roi_end[self.Plane.X]

    @property
    def roiendy(self):
        """."""
        return self._roi_end[self.Plane.Y]

    @property
    def roiprojx(self):
        """."""
        return self._roi_proj[self.Plane.X].copy()

    @property
    def roiprojy(self):
        """."""
        return self._roi_proj[self.Plane.Y].copy()

    @property
    def roiaxisx(self):
        """."""
        return self._roi_axis[self.Plane.X].copy()

    @property
    def roiaxisy(self):
        """."""
        return self._roi_axis[self.Plane.Y].copy()

    @property
    def roigaussx(self):
        """."""
        return self._roi_gauss[self.Plane.X].copy()

    @property
    def roigaussy(self):
        """."""
        return self._roi_gauss[self.Plane.Y].copy()

    @property
    def background(self):
        """."""
        return self._background.copy().ravel()

    @background.setter
    def background(self, val):
        """."""
        if not isinstance(val, _np.ndarray):
            _log.error('Could not set background')
            return
        img = self._adjust_image_dimensions(np.array(val, dtype=float))
        if img is None:
            _log.error('Could not set background')
            return
        self._background = img
        self.run_callbacks('Background-RB', self.background)

    @property
    def usebackground(self):
        """."""
        return self._background_use

    @usebackground.setter
    def usebackground(self, val):
        """."""
        self._background_use = bool(val)
        self.run_callbacks('BgUse-Sts', val)

    @property
    def beamcenterx(self):
        """."""
        return self._beam_params[self.Plane.X][self.FitParams.Cen]

    @property
    def beamcentery(self):
        """."""
        return self._beam_params[self.Plane.Y][self.FitParams.Cen]

    @property
    def beamsizex(self):
        """."""
        return abs(self._beam_params[self.Plane.X][self.FitParams.Sig])

    @property
    def beamsizey(self):
        """."""
        return abs(self._beam_params[self.Plane.Y][self.FitParams.Sig])

    @property
    def beamamplx(self):
        """."""
        return self._beam_params[self.Plane.X][self.FitParams.Amp]

    @property
    def beamamply(self):
        """."""
        return self._beam_params[self.Plane.Y][self.FitParams.Amp]

    @property
    def beamoffsetx(self):
        """."""
        return self._beam_params[self.Plane.X][self.FitParams.Off]

    @property
    def beamoffsety(self):
        """."""
        return self._beam_params[self.Plane.Y][self.FitParams.Off]

    @property
    def px2mmscalex(self):
        """."""
        return self._conv_scale[self.Plane.X]

    @px2mmscalex.setter
    def px2mmscalex(self, val):
        """."""
        if val != 0:
            self._conv_scale[self.Plane.X] = val
            self.run_callbacks('Px2mmScaleX-RB', val)
        else:
            _log.error('Could not set px2mmscaley')

    @property
    def px2mmscaley(self):
        """."""
        return self._conv_scale[self.Plane.Y]

    @px2mmscaley.setter
    def px2mmscaley(self, val):
        """."""
        if val != 0:
            self._conv_scale[self.Plane.Y] = val
            self.run_callbacks('Px2mmScaleY-RB', val)
        else:
            _log.error('Could not set px2mmscaley')

    @property
    def px2mmautocenter(self):
        """."""
        return self._conv_autocenter

    @px2mmautocenter.setter
    def px2mmautocenter(self, val):
        """."""
        self._conv_autocenter = bool(val)
        self.run_callbacks('Px2mmAutoCenter-Sts', val)

    @property
    def px2mmcenterx(self):
        """."""
        return self._conv_cen[self.Plane.X]

    @px2mmcenterx.setter
    def px2mmcenterx(self, val):
        """."""
        val = int(val)
        if 0 <= val < self._image.shape[self.Plane.X]:
            self._conv_cen[self.Plane.X] = val
            self.run_callbacks('Px2mmCenterX-RB', val)
        else:
            _log.error('Could not set px2mmcenterx')

    @property
    def px2mmcentery(self):
        """."""
        return self._conv_cen[self.Plane.Y]

    @px2mmcentery.setter
    def px2mmcentery(self, val):
        """."""
        val = int(val)
        if 0 <= val < self._image.shape[self.Plane.Y]:
            self._conv_cen[self.Plane.Y] = val
            self.run_callbacks('Px2mmCenterY-RB', val)
        else:
            _log.error('Could not set px2mmcentery')

    @property
    def beamcentermmx(self):
        """."""
        val = self.beamcenterx - self._conv_cen[self.Plane.X]
        val *= self._conv_scale[self.Plane.X]
        return val

    @property
    def beamcentermmy(self):
        """."""
        # Inverted due image origin in Pxls:
        val = self._conv_cen[self.Plane.Y] - self.beamcentery
        val *= self._conv_scale[self.Plane.Y]
        return val

    @property
    def beamsizemmx(self):
        """."""
        return self.beamsizex * self._conv_scale[self.Plane.X]

    @property
    def beamsizemmy(self):
        """."""
        return self.beamsizey * self._conv_scale[self.Plane.Y]

    def cmd_reset_buffer(self, *args):
        """Schedule reset Buffer in next update."""
        _ = args
        self._reset_buffer = True
        self.run_callbacks('BufferSize-Mon', 0)

    def _process_image(self, image):
        """."""
        image = self._adjust_image_dimensions(image)
        image = _np.array(image, dtype=float)
        if image is None:
            _log.error('Image is None')
            return

        # check whether to reset or not the buffer
        imgs = self._images
        self._reset_buffer |= bool(imgs) and (imgs[-1].shape != image.shape)
        if self._reset_buffer:
            self._images = [image, ]
            self._reset_buffer = False
        else:
            self._images.append(image)
            self._images = self._images[-self._nr_averages:]
        buf_size = self.buffer_size
        self.run_callbacks('BufferSize-Mon', buf_size)

        # calculate average of the images
        if buf_size > 1:
            image = self._images[0].copy()
            for img in self._images[1:]:  # faster than using numpy.mean
                image += img
            image /= len(self._images)

        if self._background_use and self._background.shape == image.shape:
            image -= self._background
            b = _np.where(image < 0)
            image[b] = 0
        else:
            self.usebackground = False

        if self._crop_use:
            boo = image > self._crop[self.CropIdx.High]
            image[boo] = self._crop[self.CropIdx.High]
            boo = image < self._crop[self.CropIdx.Low]
            image[boo] = self._crop[self.CropIdx.Low]

        if self._flip[self.Plane.X]:
            image = _np.flip(image, axis=self.Plane.X)
        if self._flip[self.Plane.Y]:
            image = _np.flip(image, axis=self.Plane.Y)

        self._image = image
        self.run_callbacks('Image-RB', self.image)
        self._update_roi()
        axisx = self._roi_axis[self.Plane.X]
        projx = self._roi_proj[self.Plane.X]
        axisy = self._roi_axis[self.Plane.Y]
        projy = self._roi_proj[self.Plane.Y]
        if self._method == self.Method.Moments:
            parx = self._calc_moments(axisx, projx)
            pary = self._calc_moments(axisy, projy)
        else:
            parx = self._fit_gaussian(
                axisx, projx, self._beam_params[self.Plane.X])
            pary = self._fit_gaussian(
                axisy, projy, self._beam_params[self.Plane.Y])
        self._roi_gauss[self.Plane.X] = self._gaussian(axisx, *parx)
        self._roi_gauss[self.Plane.Y] = self._gaussian(axisy, *pary)
        self._beam_params[self.Plane.X] = parx
        self._beam_params[self.Plane.Y] = pary
        self.run_callbacks('ROIGaussFitX-Mon', self._roi_gauss[self.Plane.X])
        self.run_callbacks('ROIGaussFitY-Mon', self._roi_gauss[self.Plane.Y])
        self.run_callbacks('BeamCenterX-Mon', self.beamcenterx)
        self.run_callbacks('BeamCenterY-Mon', self.beamcentery)
        self.run_callbacks('BeamSizeX-Mon', self.beamsizex)
        self.run_callbacks('BeamSizeY-Mon', self.beamsizey)
        self.run_callbacks('BeamCentermmX-Mon', self.beamcentermmx)
        self.run_callbacks('BeamCentermmY-Mon', self.beamcentermmy)
        self.run_callbacks('BeamSizemmX-Mon', self.beamsizemmx)
        self.run_callbacks('BeamSizemmY-Mon', self.beamsizemmy)
        self.run_callbacks('BeamAmplX-Mon', self.beamamplx)
        self.run_callbacks('BeamAmplY-Mon', self.beamamply)
        self.run_callbacks('BeamOffsetX-Mon', self.beamoffsetx)
        self.run_callbacks('BeamOffsetY-Mon', self.beamoffsety)

    def _adjust_image_dimensions(self, img):
        """."""
        if len(img.shape) == 1:
            if self._width <= 1:
                _log.error('Invalid value for Width.')
                return None
            try:
                if self._reading_order == self.ReadingOrder.CLike:
                    img = img.reshape((-1, self._width), order='C')
                else:
                    img = img.reshape((self._width, -1), order='F')
            except ValueError:
                _log.error('Problem reshaping image.')
                return None
        return img

    def _update_roi(self):
        """."""
        image = self._image
        axis_x = _np.arange(image.shape[self.Plane.X])
        axis_y = _np.arange(image.shape[self.Plane.Y])

        if self._conv_autocenter:
            self.px2mmcenterx = image.shape[self.Plane.X]//2
            self.px2mmcentery = image.shape[self.Plane.Y]//2

        if self._roi_autocenter:
            proj_x = image.sum(axis=0)
            proj_y = image.sum(axis=1)
            parx = self._calc_moments(axis_x, proj_x)
            pary = self._calc_moments(axis_y, proj_y)
            if not any(_np.isnan(parx+pary)):
                self.roicenterx = int(parx[self.FitParams.Cen])
                self.roicentery = int(pary[self.FitParams.Cen])
            else:
                _log.error('Some fitted params are NaN.')

        strtx = self._roi_cen[self.Plane.X] - self._roi_size[self.Plane.X]
        endx = self._roi_cen[self.Plane.X] + self._roi_size[self.Plane.X]
        strty = self._roi_cen[self.Plane.Y] - self._roi_size[self.Plane.Y]
        endy = self._roi_cen[self.Plane.Y] + self._roi_size[self.Plane.Y]
        strtx, strty = max(strtx, 0), max(strty, 0)
        endx = min(endx, image.shape[self.Plane.X])
        endy = min(endy, image.shape[self.Plane.Y])

        image = image[strty:endy, strtx:endx]
        self._roi_proj[self.Plane.X] = image.sum(axis=self.Plane.Y)
        self._roi_proj[self.Plane.Y] = image.sum(axis=self.Plane.X)
        self._roi_axis[self.Plane.X] = axis_x[strtx:endx]
        self._roi_axis[self.Plane.Y] = axis_y[strty:endy]
        self._roi_start[self.Plane.X] = strtx
        self._roi_start[self.Plane.Y] = strty
        self._roi_end[self.Plane.X] = endx
        self._roi_end[self.Plane.Y] = endy
        self.run_callbacks('ROIProjX-Mon', self._roi_proj[self.Plane.X])
        self.run_callbacks('ROIProjY-Mon', self._roi_proj[self.Plane.Y])
        self.run_callbacks('ROIAxisX-Mon', self._roi_axis[self.Plane.X])
        self.run_callbacks('ROIAxisY-Mon', self._roi_axis[self.Plane.Y])
        self.run_callbacks('ROIStartX-Mon', self._roi_start[self.Plane.X])
        self.run_callbacks('ROIStartY-Mon', self._roi_start[self.Plane.Y])
        self.run_callbacks('ROIEndX-Mon', self._roi_end[self.Plane.X])
        self.run_callbacks('ROIEndY-Mon', self._roi_end[self.Plane.Y])

    @classmethod
    def _calc_moments(cls, axis, proj):
        """."""
        ret = [0, 0, 0, 0]
        y0 = _np.min(proj)
        proj = proj - y0
        amp = _np.amax(proj)
        dx = axis[1]-axis[0]
        norm = _np.trapz(proj, dx=dx)
        praxis = proj*axis
        cen = _np.trapz(praxis, dx=dx)/norm
        sec = _np.trapz(praxis*axis, dx=dx)/norm
        std = _np.sqrt(sec - cen*cen)
        ret[cls.FitParams.Amp] = amp
        ret[cls.FitParams.Cen] = cen
        ret[cls.FitParams.Sig] = std
        ret[cls.FitParams.Off] = y0
        return ret

    @classmethod
    def _gaussian(cls, x, *args):
        """."""
        mu = args[cls.FitParams.Cen]
        sigma = args[cls.FitParams.Sig]
        y0 = args[cls.FitParams.Off]
        amp = args[cls.FitParams.Amp]
        x = x - mu
        return amp*_np.exp(-x*x/(2.0*sigma*sigma))+y0

    @classmethod
    def _fit_gaussian(cls, x, y, par=None):
        """."""
        if par is None:
            par = cls._calc_moments(x, y)
        try:
            par, _ = curve_fit(cls._gaussian, x, y, par)
        except Exception:
            _log.error('Could not fit gaussian.')
        return par



"""Module to deal with orbit acquisition."""
import logging as _log
import numpy as _np

from ..epics import PV as _PV
from ..diagbeam.bpm.csdev import Const as _csbpm
from ..timesys.csdev import Const as _TIConst
from ..search import HLTimeSearch as _HLTimesearch
from ..envars import VACA_PREFIX as LL_PREF

from .base_class import BaseTimingConfig as _BaseTimingConfig

TIMEOUT = 0.05


class BPM(_BaseTimingConfig):
    """."""

    def __init__(self, name, callback=None):
        """."""
        super().__init__(name[:2], callback)
        self._name = name
        self._orb_conv_unit = self._csorb.ORBIT_CONVERSION_UNIT
        pvpref = LL_PREF + ('-' if LL_PREF else '') + self._name + ':'
        opt = {'connection_timeout': TIMEOUT}
        self._poskx = _PV(pvpref + 'PosKx-RB', **opt)
        self._posky = _PV(pvpref + 'PosKy-RB', **opt)
        self._ksum = _PV(pvpref + 'PosKsum-RB', **opt)
        self._spposx = _PV(pvpref + 'SPPosX-Mon', **opt)
        self._spposy = _PV(pvpref + 'SPPosY-Mon', **opt)
        self._spsum = _PV(pvpref + 'SPSum-Mon', **opt)
        self._polyx = _PV(pvpref + 'GEN_PolyXArrayCoeff-RB', **opt)
        self._polyy = _PV(pvpref + 'GEN_PolyYArrayCoeff-RB', **opt)
        opt['auto_monitor'] = False
        self._arraya = _PV(pvpref + 'GEN_AArrayData', **opt)
        self._arrayb = _PV(pvpref + 'GEN_BArrayData', **opt)
        self._arrayc = _PV(pvpref + 'GEN_CArrayData', **opt)
        self._arrayd = _PV(pvpref + 'GEN_DArrayData', **opt)
        self._arrayx = _PV(pvpref + 'GEN_XArrayData', **opt)
        self._arrayy = _PV(pvpref + 'GEN_YArrayData', **opt)
        self._arrays = _PV(pvpref + 'GEN_SUMArrayData', **opt)
        opt.pop('auto_monitor')
        self._offsetx = _PV(pvpref + 'PosXOffset-RB', **opt)
        self._offsety = _PV(pvpref + 'PosYOffset-RB', **opt)
        self._config_ok_vals = {
            'asyn.ENBL': _csbpm.EnblTyp.Enable,
            'SwMode': _csbpm.SwModes.switching,
            'ACQBPMMode': _csbpm.OpModes.MultiBunch,
            'ACQChannel': _csbpm.AcqChan.ADC,
            # 'ACQNrShots': 1,
            'ACQShots': 1,
            # 'ACQTriggerHwDly': 0.0,  # NOTE: leave this property commented
            'ACQUpdateTime': 0.001,
            # 'ACQNrSamplesPre': 0,
            'ACQSamplesPre': 0,
            # 'ACQNrSamplesPost': 200,
            'ACQSamplesPost': 360,
            # 'ACQCtrl': _csbpm.AcqEvents.Stop,
            'ACQTriggerEvent': _csbpm.AcqEvents.Stop,
            # 'ACQTriggerType': _csbpm.AcqTrigTyp.External,
            'ACQTrigger': _csbpm.AcqTrigTyp.External,
            'ACQTriggerRep': _csbpm.AcqRepeat.Normal,
            # 'ACQTriggerDataChan': _csbpm.AcqChan.Monit1,
            'ACQDataTrigChan': _csbpm.AcqChan.ADC,
            'ACQTriggerDataSel': _csbpm.AcqDataTyp.A,
            'ACQTriggerDataThres': 1,
            'ACQTriggerDataPol': _csbpm.Polarity.Positive,
            'ACQTriggerDataHyst': 0,
            'TbtTagEn': _csbpm.EnbldDsbld.disabled,  # Enable TbT sync Timing
            'SwTagEn': _csbpm.EnbldDsbld.disabled,  # Enable FOFB sync Timing
            'Monit1TagEn': _csbpm.EnbldDsbld.disabled,
            'MonitTagEn': _csbpm.EnbldDsbld.disabled,
            'TbtDataMaskEn': _csbpm.EnbldDsbld.disabled,  # Enable use of mask
            'TbtDataMaskSamplesBeg': 0,
            'TbtDataMaskSamplesEnd': 0,
            'XYPosCal': _csbpm.EnbldDsbld.enabled,
            'SUMPosCal': _csbpm.EnbldDsbld.enabled}
        pvs = {
            'asyn.ENBL': 'asyn.ENBL',
            'SwMode': 'SwMode-Sel',
            'ACQBPMMode': 'ACQBPMMode-Sel',
            'ACQChannel': 'ACQChannel-Sel',
            # 'ACQNrShots': 'ACQNrShots-SP',
            'ACQShots': 'ACQShots-SP',
            # 'ACQTriggerHwDly': 'ACQTriggerHwDly-SP',
            'ACQUpdateTime': 'ACQUpdateTime-SP',
            # 'ACQNrSamplesPre': 'ACQNrSamplesPre-SP',
            'ACQSamplesPre': 'ACQSamplesPre-SP',
            # 'ACQNrSamplesPost': 'ACQNrSamplesPost-SP',
            'ACQSamplesPost': 'ACQSamplesPost-SP',
            # 'ACQCtrl': 'ACQCtrl-Sel',
            'ACQTriggerEvent': 'ACQTriggerEvent-Sel',
            # 'ACQTriggerType': 'ACQTriggerType-Sel',
            'ACQTrigger': 'ACQTrigger-Sel',
            'ACQTriggerRep': 'ACQTriggerRep-Sel',
            # 'ACQTriggerDataChan': 'ACQTriggerDataChan-Sel',
            'ACQDataTrigChan': 'ACQDataTrigChan-Sel',
            # 'ACQTriggerDataSel': 'ACQTriggerDataSel-Sel',
            'ACQTriggerDataSel': 'ACQTriggerDataSel-SP',
            'ACQTriggerDataThres': 'ACQTriggerDataThres-SP',
            'ACQTriggerDataPol': 'ACQTriggerDataPol-Sel',
            'ACQTriggerDataHyst': 'ACQTriggerDataHyst-SP',
            'TbtTagEn': 'TbtTagEn-Sel',  # Enable TbT sync with timing
            'SwTagEn': 'SwTagEn-Sel',  # Enable FOFB sync with timing
            'Monit1TagEn': 'Monit1TagEn-Sel',  # Enable Monit1 sync with timing
            'MonitTagEn': 'MonitTagEn-Sel',  # Enable Monit sync with timing
            'TbtDataMaskEn': 'TbtDataMaskEn-Sel',  # Enable use of mask
            'TbtDataMaskSamplesBeg': 'TbtDataMaskSamplesBeg-SP',
            'TbtDataMaskSamplesEnd': 'TbtDataMaskSamplesEnd-SP',
            'XYPosCal': 'XYPosCal-Sel',
            'SUMPosCal': 'SUMPosCal-Sel'}

        self._config_pvs_sp = {
            k: _PV(pvpref + v, **opt) for k, v in pvs.items()}
        pvs = {
            'asyn.ENBL': 'asyn.ENBL',
            'asyn.CNCT': 'asyn.CNCT',
            'INFOClkFreq': 'INFOClkFreq-RB',
            'INFOHarmonicNumber': 'INFOHarmonicNumber-RB',
            'INFOTBTRate': 'INFOTBTRate-RB',
            'INFOFOFBRate': 'INFOFOFBRate-RB',
            'INFOMONITRate': 'INFOMONITRate-RB',
            'INFOMONIT1Rate': 'INFOMONIT1Rate-RB',
            'SwMode': 'SwMode-Sts',
            'ACQBPMMode': 'ACQBPMMode-Sts',
            'ACQChannel': 'ACQChannel-Sts',
            # 'ACQNrShots': 'ACQNrShots-RB',
            'ACQShots': 'ACQShots-RB',
            # 'ACQTriggerHwDly': 'ACQTriggerHwDly-RB',
            'ACQUpdateTime': 'ACQUpdateTime-RB',
            # 'ACQNrSamplesPre': 'ACQNrSamplesPre-RB',
            'ACQSamplesPre': 'ACQSamplesPre-RB',
            # 'ACQNrSamplesPost': 'ACQNrSamplesPost-RB',
            'ACQSamplesPost': 'ACQSamplesPost-RB',
            # 'ACQCtrl': 'ACQCtrl-Sts',
            'ACQTriggerEvent': 'ACQTriggerEvent-Sts',
            # 'ACQStatus': 'ACQStatus-Mon',
            'ACQStatus': 'ACQStatus-Sts',
            # 'ACQTriggerType': 'ACQTriggerType-Sts',
            'ACQTrigger': 'ACQTrigger-Sts',
            'ACQTriggerRep': 'ACQTriggerRep-Sts',
            # 'ACQTriggerDataChan': 'ACQTriggerDataChan-Sts',
            'ACQDataTrigChan': 'ACQDataTrigChan-Sts',
            # 'ACQTriggerDataSel': 'ACQTriggerDataSel-Sts',
            'ACQTriggerDataSel': 'ACQTriggerDataSel-RB',
            'ACQTriggerDataThres': 'ACQTriggerDataThres-RB',
            'ACQTriggerDataPol': 'ACQTriggerDataPol-Sts',
            'ACQTriggerDataHyst': 'ACQTriggerDataHyst-RB',
            'TbtTagEn': 'TbtTagEn-Sts',
            'SwTagEn': 'SwTagEn-Sts',
            'Monit1TagEn': 'Monit1TagEn-Sts',
            'MonitTagEn': 'MonitTagEn-Sts',
            'TbtDataMaskEn': 'TbtDataMaskEn-Sts',
            'TbtDataMaskSamplesBeg': 'TbtDataMaskSamplesBeg-RB',
            'TbtDataMaskSamplesEnd': 'TbtDataMaskSamplesEnd-RB',
            'XYPosCal': 'XYPosCal-Sts',
            'SUMPosCal': 'SUMPosCal-Sts'}
        self._config_pvs_rb = {
            k: _PV(pvpref + v, **opt) for k, v in pvs.items()}

    @property
    def name(self):
        """."""
        return self._name

    @property
    def connected(self):
        """."""
        conn = super().connected
        pvs = (
            self._spposx, self._spposy, self._spsum,
            self._arrayx, self._arrayy, self._arrays,
            self._offsetx, self._offsety,
            self._polyx, self._polyx,
            self._arraya, self._arrayb, self._arrayc, self._arrayd,
            self._poskx, self._posky, self._ksum)
        for pvobj in pvs:
            if not pvobj.connected:
                _log.debug('NOT CONN: ' + pvobj.pvname)
            conn &= pvobj.connected
        return conn

    @property
    def is_ok(self):
        """."""
        if not super().is_ok:
            return False

        pvobj = self._config_pvs_rb['ACQStatus']
        stts = _csbpm.AcqStates
        okay = pvobj.value not in {
            stts.Error, stts.No_Memory, stts.Too_Few_Samples,
            stts.Too_Many_Samples, stts.Acq_Overflow}

        if self._config_ok_vals['ACQTriggerEvent'] == _csbpm.AcqEvents.Start:
            okay &= pvobj.value not in {stts.Idle, stts.Aborted}
        else:
            okay &= pvobj.value not in {
                stts.Waiting, stts.External_Trig, stts.Data_Trig,
                stts.Software_Trig, stts.Acquiring}
        if not okay:
            msg = 'ERR: Error in {0:s}'.format(pvobj.pvname)
            self.run_callbacks('Log-Mon', msg)
            _log.warning(msg[5:])
        return okay

    @property
    def state(self):
        """."""
        pvobj = self._config_pvs_rb['asyn.ENBL']
        if pvobj.connected:
            return pvobj.value == _csbpm.EnblTyp.Enable
        return False

    @state.setter
    def state(self, boo):
        """."""
        val = _csbpm.EnblTyp.Enable if boo else _csbpm.EnblTyp.Disable
        pvobj = self._config_pvs_sp['asyn.ENBL']
        self._config_ok_vals['asyn.ENBL'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def switching_mode(self):
        """."""
        pvobj = self._config_pvs_rb['SwMode']
        if pvobj.connected:
            return pvobj.value
        return None

    @switching_mode.setter
    def switching_mode(self, val):
        """."""
        pvobj = self._config_pvs_sp['SwMode']
        self._config_ok_vals['SwMode'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def adcfreq(self):
        """."""
        defv = 218446014.0 if self._csorb.acc == 'BO' else 220870069.0
        pvobj = self._config_pvs_rb['INFOClkFreq']
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def tbtrate(self):
        """."""
        defv = 362 if self._csorb.acc == 'BO' else 382
        pvobj = self._config_pvs_rb['INFOTBTRate']
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def tbtperiod(self):
        """."""
        return self.tbtrate / self.adcfreq

    @property
    def fofbrate(self):
        """."""
        defv = (362 if self._csorb.acc == 'BO' else 382) * 24
        pvobj = self._config_pvs_rb['INFOFOFBRate']
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def fofbperiod(self):
        """."""
        return self.fofbrate / self.adcfreq

    @property
    def monitrate(self):
        """."""
        defv = (362 if self._csorb.acc == 'BO' else 382) * 59904
        pvobj = self._config_pvs_rb['INFOMONITRate']
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def monitperiod(self):
        """."""
        return self.monitrate / self.adcfreq

    @property
    def monit1rate(self):
        """."""
        defv = (362 if self._csorb.acc == 'BO' else 382) * 603
        pvobj = self._config_pvs_rb['INFOMONIT1Rate']
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def monit1period(self):
        """."""
        return self.monit1rate / self.adcfreq

    @property
    def poskx(self):
        """."""
        defv = 1
        pvobj = self._poskx
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def posky(self):
        """."""
        defv = 1
        pvobj = self._posky
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def polyx(self):
        """."""
        pvobj = self._polyx
        if pvobj.connected:
            val = pvobj.value
            if val is not None:
                return val
        defv = _np.zeros(15, dtype=float)
        defv[0] = 1
        return defv

    @property
    def polyy(self):
        """."""
        pvobj = self._polyy
        if pvobj.connected:
            val = pvobj.value
            if val is not None:
                return val
        defv = _np.zeros(15, dtype=float)
        defv[0] = 1
        return defv

    @property
    def ksum(self):
        """."""
        defv = 1
        pvobj = self._ksum
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @property
    def mode(self):
        """."""
        pvobj = self._config_pvs_rb['ACQBPMMode']
        return pvobj.value if pvobj.connected else _csbpm.OpModes.MultiBunch

    @mode.setter
    def mode(self, mode):
        """."""
        pvobj = self._config_pvs_sp['ACQBPMMode']
        self._config_ok_vals['ACQBPMMode'] = mode
        if self.put_enable and pvobj.connected:
            pvobj.value = mode

    @property
    def spposx(self):
        """."""
        pvobj = self._spposx
        val = pvobj.value if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit*val

    @property
    def spposy(self):
        """."""
        pvobj = self._spposy
        val = pvobj.value if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit*val

    @property
    def spsum(self):
        """."""
        pvobj = self._spsum
        val = pvobj.value if pvobj.connected else None
        if val is not None:
            return val

    @property
    def arraya(self):
        """."""
        pvobj = self._arraya
        return pvobj.get() if pvobj.connected else None

    @property
    def arrayb(self):
        """."""
        pvobj = self._arrayb
        return pvobj.get() if pvobj.connected else None

    @property
    def arrayc(self):
        """."""
        pvobj = self._arrayc
        return pvobj.get() if pvobj.connected else None

    @property
    def arrayd(self):
        """."""
        pvobj = self._arrayd
        return pvobj.get() if pvobj.connected else None

    @property
    def mtposx(self):
        """."""
        pvobj = self._arrayx
        val = pvobj.get() if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit*val

    @property
    def mtposy(self):
        """."""
        pvobj = self._arrayy
        val = pvobj.get() if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit*val

    @property
    def mtsum(self):
        """."""
        pvobj = self._arrays
        val = pvobj.get() if pvobj.connected else None
        if val is not None:
            return val

    @property
    def offsetx(self):
        """."""
        pvobj = self._offsetx
        val = pvobj.value if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit*val

    @property
    def offsety(self):
        """."""
        pvobj = self._offsety
        val = pvobj.value if pvobj.connected else None
        if val is not None:
            return self._orb_conv_unit*val

    @property
    def ctrl(self):
        """."""
        # pvobj = self._config_pvs_rb['ACQCtrl']
        pvobj = self._config_pvs_rb['ACQTriggerEvent']
        return pvobj.value if pvobj.connected else None

    @ctrl.setter
    def ctrl(self, val):
        """."""
        # pvobj = self._config_pvs_sp['ACQCtrl']
        pvobj = self._config_pvs_sp['ACQTriggerEvent']
        # self._config_ok_vals['ACQCtrl'] = val
        self._config_ok_vals['ACQTriggerEvent'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_type(self):
        """."""
        pvobj = self._config_pvs_rb['ACQChannel']
        return pvobj.value if pvobj.connected else None

    @acq_type.setter
    def acq_type(self, val):
        """."""
        pvobj = self._config_pvs_sp['ACQChannel']
        self._config_ok_vals['ACQChannel'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_trigger(self):
        """."""
        # pvobj = self._config_pvs_rb['ACQTriggerType']
        pvobj = self._config_pvs_rb['ACQTrigger']
        return pvobj.value if pvobj.connected else None

    @acq_trigger.setter
    def acq_trigger(self, val):
        """."""
        # pvobj = self._config_pvs_sp['ACQTriggerType']
        pvobj = self._config_pvs_sp['ACQTrigger']
        # self._config_ok_vals['ACQTriggerType'] = val
        self._config_ok_vals['ACQTrigger'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_repeat(self):
        """."""
        pvobj = self._config_pvs_rb['ACQTriggerRep']
        return pvobj.value if pvobj.connected else None

    @acq_repeat.setter
    def acq_repeat(self, val):
        """."""
        pvobj = self._config_pvs_sp['ACQTriggerRep']
        self._config_ok_vals['ACQTriggerRep'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_trig_datatype(self):
        """."""
        # pvobj = self._config_pvs_rb['ACQTriggerDataChan']
        pvobj = self._config_pvs_rb['ACQDataTrigChan']
        return pvobj.value if pvobj.connected else None

    @acq_trig_datatype.setter
    def acq_trig_datatype(self, val):
        """."""
        # pvobj = self._config_pvs_sp['ACQTriggerDataChan']
        pvobj = self._config_pvs_sp['ACQDataTrigChan']
        # self._config_ok_vals['ACQTriggerDataChan'] = val
        self._config_ok_vals['ACQDataTrigChan'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_trig_datasel(self):
        """."""
        pvobj = self._config_pvs_rb['ACQTriggerDataSel']
        return pvobj.value if pvobj.connected else None

    @acq_trig_datasel.setter
    def acq_trig_datasel(self, val):
        """."""
        pvobj = self._config_pvs_sp['ACQTriggerDataSel']
        self._config_ok_vals['ACQTriggerDataSel'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_trig_datathres(self):
        """."""
        pvobj = self._config_pvs_rb['ACQTriggerDataThres']
        return pvobj.value if pvobj.connected else None

    @acq_trig_datathres.setter
    def acq_trig_datathres(self, val):
        """."""
        pvobj = self._config_pvs_sp['ACQTriggerDataThres']
        self._config_ok_vals['ACQTriggerDataThres'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_trig_datahyst(self):
        """."""
        pvobj = self._config_pvs_rb['ACQTriggerDataHyst']
        return pvobj.value if pvobj.connected else None

    @acq_trig_datahyst.setter
    def acq_trig_datahyst(self, val):
        """."""
        pvobj = self._config_pvs_sp['ACQTriggerDataHyst']
        self._config_ok_vals['ACQTriggerDataHyst'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def acq_trig_datapol(self):
        """."""
        pvobj = self._config_pvs_rb['ACQTriggerDataPol']
        return pvobj.value if pvobj.connected else None

    @acq_trig_datapol.setter
    def acq_trig_datapol(self, val):
        """."""
        pvobj = self._config_pvs_sp['ACQTriggerDataPol']
        self._config_ok_vals['ACQTriggerDataPol'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def tbt_sync_enbl(self):
        """."""
        pvobj = self._config_pvs_rb['TbtTagEn']
        return pvobj.value if pvobj.connected else None

    @tbt_sync_enbl.setter
    def tbt_sync_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp['TbtTagEn']
        self._config_ok_vals['TbtTagEn'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def fofb_sync_enbl(self):
        """."""
        pvobj = self._config_pvs_rb['SwTagEn']
        return pvobj.value if pvobj.connected else None

    @fofb_sync_enbl.setter
    def fofb_sync_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp['SwTagEn']
        self._config_ok_vals['SwTagEn'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def monit1_sync_enbl(self):
        """."""
        pvobj = self._config_pvs_rb['Monit1TagEn']
        return pvobj.value if pvobj.connected else None

    @monit1_sync_enbl.setter
    def monit1_sync_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp['Monit1TagEn']
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def monit_sync_enbl(self):
        """."""
        pvobj = self._config_pvs_rb['MonitTagEn']
        return pvobj.value if pvobj.connected else None

    @monit_sync_enbl.setter
    def monit_sync_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp['MonitTagEn']
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def tbt_mask_enbl(self):
        """."""
        pvobj = self._config_pvs_rb['TbtDataMaskEn']
        return pvobj.value if pvobj.connected else None

    @tbt_mask_enbl.setter
    def tbt_mask_enbl(self, val):
        """."""
        pvobj = self._config_pvs_sp['TbtDataMaskEn']
        self._config_ok_vals['TbtDataMaskEn'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def tbt_mask_begin(self):
        """."""
        pvobj = self._config_pvs_rb['TbtDataMaskSamplesBeg']
        return pvobj.value if pvobj.connected else None

    @tbt_mask_begin.setter
    def tbt_mask_begin(self, val):
        """."""
        pvobj = self._config_pvs_sp['TbtDataMaskSamplesBeg']
        self._config_ok_vals['TbtDataMaskSamplesBeg'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def tbt_mask_end(self):
        """."""
        pvobj = self._config_pvs_rb['TbtDataMaskSamplesEnd']
        return pvobj.value if pvobj.connected else None

    @tbt_mask_end.setter
    def tbt_mask_end(self, val):
        """."""
        pvobj = self._config_pvs_sp['TbtDataMaskSamplesEnd']
        self._config_ok_vals['TbtDataMaskSamplesEnd'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def polycal(self):
        """."""
        pvobj = self._config_pvs_rb['XYPosCal']
        return pvobj.value if pvobj.connected else None

    @polycal.setter
    def polycal(self, val):
        """."""
        val = _csbpm.EnbldDsbld.enabled if val else _csbpm.EnbldDsbld.disabled
        pv1 = self._config_pvs_sp['XYPosCal']
        pv2 = self._config_pvs_sp['SUMPosCal']
        self._config_ok_vals['XYPosCal'] = val
        self._config_ok_vals['SUMPosCal'] = val
        if self.put_enable and pv1.connected:
            pv1.put(val, wait=False)
        if self.put_enable and pv2.connected:
            pv2.put(val, wait=False)

    @property
    def nrsamplespost(self):
        """."""
        # pvobj = self._config_pvs_rb['ACQNrSamplesPost']
        pvobj = self._config_pvs_rb['ACQSamplesPost']
        return pvobj.value if pvobj.connected else None

    @nrsamplespost.setter
    def nrsamplespost(self, val):
        """."""
        # pvobj = self._config_pvs_sp['ACQNrSamplesPost']
        pvobj = self._config_pvs_sp['ACQSamplesPost']
        # self._config_ok_vals['ACQNrSamplesPost'] = val
        self._config_ok_vals['ACQSamplesPost'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def nrsamplespre(self):
        """."""
        # pvobj = self._config_pvs_rb['ACQNrSamplesPre']
        pvobj = self._config_pvs_rb['ACQSamplesPre']
        return pvobj.value if pvobj.connected else None

    @nrsamplespre.setter
    def nrsamplespre(self, val):
        """."""
        # pvobj = self._config_pvs_sp['ACQNrSamplesPre']
        pvobj = self._config_pvs_sp['ACQSamplesPre']
        # self._config_ok_vals['ACQNrSamplesPre'] = val
        self._config_ok_vals['ACQSamplesPre'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def nrshots(self):
        """."""
        # pvobj = self._config_pvs_rb['ACQNrShots']
        pvobj = self._config_pvs_rb['ACQShots']
        return pvobj.value if pvobj.connected else None

    @nrshots.setter
    def nrshots(self, val):
        """."""
        # pvobj = self._config_pvs_sp['ACQNrShots']
        pvobj = self._config_pvs_sp['ACQShots']
        # self._config_ok_vals['ACQNrShots'] = val
        self._config_ok_vals['ACQShots'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    def calc_sp_multiturn_pos(self, **kwargs):
        """."""
        nturns = kwargs.get('nturns', 1)
        refx = kwargs.get('refx', 0.0)
        refy = kwargs.get('refy', 0.0)
        refsum = kwargs.get('refsum', 0.0)
        maskbeg = kwargs.get('maskbeg', 0)
        maskend = kwargs.get('maskend', 0)
        bgd = kwargs.get('bg', dict())

        wsize = self.tbtrate
        maskbeg = min(maskbeg, wsize - 2)
        maskend = min(maskend, wsize - maskbeg - 2)
        mask = slice(maskbeg, wsize - maskend)

        # NOTE: I have to invert array B with C here because of the way
        # the ADCSWAP rate works. Fixed in 2020/07/01 after talking to
        # Daniel Tavares.
        vals = {
            'A': self.arraya, 'C': self.arrayb,
            'B': self.arrayc, 'D': self.arrayd}
        siz = None
        for key, val in vals.items():
            if val is None or val.size == 0:
                siz = 0
                break
            nzrs = val.size
            siz = nzrs if siz is None else min(siz, nzrs)
            if bgd and bgd[key].size >= val.size:
                vals[key] -= bgd[key][:val.size]

        x_cal = _np.full(nturns, refx)
        y_cal = _np.full(nturns, refy)
        s_cal = _np.full(nturns, refsum)

        # handle cases where length read is smaller than required.
        rnts = min(siz//wsize, nturns)
        if not (siz and rnts):
            return x_cal, y_cal, s_cal

        for key, val in vals.items():
            val = val[:(rnts*wsize)]
            val = val.reshape(-1, wsize)[:, mask]
            vals[key] = _np.std(val, axis=1)

        sum1, sum2 = vals['A'] + vals['C'], vals['D'] + vals['B']
        zero1 = _np.logical_not(_np.isclose(sum1, 0.0))
        zero2 = _np.logical_not(_np.isclose(sum2, 0.0))
        diff1 = (vals['A'][zero1] - vals['C'][zero1]) / sum1[zero1]
        diff2 = (vals['D'][zero2] - vals['B'][zero2]) / sum2[zero2]
        x_uncal = (diff1 + diff2) / 2
        y_uncal = (diff1 - diff2) / 2
        if self._config_ok_vals['XYPosCal'] == _csbpm.EnbldDsbld.disabled:
            x_cal[:rnts][zero1] = x_uncal * self.poskx
            y_cal[:rnts][zero2] = y_uncal * self.posky
        else:
            x_cal[:rnts][zero1], y_cal[:rnts][zero2] = self._apply_polyxy(
                x_uncal, y_uncal)
        x_cal[:rnts][zero1] *= self._orb_conv_unit
        y_cal[:rnts][zero2] *= self._orb_conv_unit
        x_cal[:rnts][zero1] -= self.offsetx or 0.0
        y_cal[:rnts][zero2] -= self.offsety or 0.0
        s_cal[:rnts] = (sum1 + sum2) * self.ksum
        return x_cal, y_cal, s_cal

    def _apply_polyxy(self, x_uncal, y_uncal):
        """."""
        x_cal = self._calc_poly(x_uncal, y_uncal, plane='x')
        y_cal = self._calc_poly(y_uncal, x_uncal, plane='y')
        return x_cal, y_cal

    def _calc_poly(self, th1, ot1, plane='x'):
        """."""
        ot2 = ot1*ot1
        ot4 = ot2*ot2
        ot6 = ot4*ot2
        ot8 = ot4*ot4
        th2 = th1*th1
        th3 = th2*th1
        th5 = th3*th2
        th7 = th5*th2
        th9 = th7*th2
        pol = self.polyx if plane == 'x' else self.polyy

        return (
            th1*(pol[0] + ot2*pol[1] + ot4*pol[2] + ot6*pol[3] + ot8*pol[4]) +
            th3*(pol[5] + ot2*pol[6] + ot4*pol[7] + ot6*pol[8]) +
            th5*(pol[9] + ot2*pol[10] + ot4*pol[11]) +
            th7*(pol[12] + ot2*pol[13]) +
            th9*pol[14])


class TimingConfig(_BaseTimingConfig):
    """."""

    def __init__(self, acc, callback=None):
        """."""
        super().__init__(acc, callback=callback)
        trig = self._csorb.trigger_acq_name
        evg = self._csorb.evg_name
        opt = {'connection_timeout': TIMEOUT}
        self._config_ok_vals = {
            'NrPulses': 1, 'State': _TIConst.TrigStates.Enbl}
        if _HLTimesearch.has_delay_type(trig):
            self._config_ok_vals['RFDelayType'] = _TIConst.TrigDlyTyp.Manual
        pref_name = LL_PREF + trig + ':'
        self._config_pvs_rb = {
            'Delay': _PV(pref_name + 'Delay-RB', **opt),
            'TotalDelay': _PV(pref_name + 'TotalDelay-Mon', **opt),
            'NrPulses': _PV(pref_name + 'NrPulses-RB', **opt),
            'Duration': _PV(pref_name + 'Duration-RB', **opt),
            'State': _PV(pref_name + 'State-Sts', **opt),
            'Injecting': _PV(LL_PREF + evg + ':InjectionEvt-Sts', **opt),
            'EGTrig': _PV('LI-01:EG-TriggerPS:status', **opt),
            }
        self._config_pvs_sp = {
            'NrPulses': _PV(pref_name + 'NrPulses-SP', **opt),
            'State': _PV(pref_name + 'State-Sel', **opt)}
        if _HLTimesearch.has_delay_type(trig):
            self._config_pvs_rb['RFDelayType'] = _PV(
                pref_name + 'RFDelayType-Sts', **opt)
            self._config_pvs_sp['RFDelayType'] = _PV(
                pref_name + 'RFDelayType-Sel', **opt)

    @property
    def injecting(self):
        """."""
        inj = bool(self._config_pvs_rb['Injecting'].value)
        eg_trig = bool(self._config_pvs_rb['EGTrig'].value)
        return inj and eg_trig

    @property
    def nrpulses(self):
        """."""
        pvobj = self._config_pvs_rb['NrPulses']
        return pvobj.value if pvobj.connected else None

    @nrpulses.setter
    def nrpulses(self, val):
        """."""
        pvobj = self._config_pvs_sp['NrPulses']
        self._config_ok_vals['NrPulses'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def duration(self):
        """."""
        pvobj = self._config_pvs_rb['Duration']
        return pvobj.value if pvobj.connected else None

    @property
    def delay(self):
        """."""
        pvobj = self._config_pvs_rb['Delay']
        return pvobj.value if pvobj.connected else None

    @property
    def totaldelay(self):
        """."""
        pvobj = self._config_pvs_rb['TotalDelay']
        return pvobj.value if pvobj.connected else None



"""PRU Controller subpackage.

This subpackage contains lower level modules and classes that are responsible
for communicating ultimately with power supplies using BSMP protocol.

Modules:

    pru.py :
        PRU library class that implements UART communication methods.
        It inkokes EthBridgePRUserial485 library methods.

    prucontroller.py :
        Implements PRUController, class used to drive the communication flow
        and store devices state mirrors.

    udc.py :
        Communication methods that apply to all devices controlled by a
        given UDC.

    prucparams.py :
        Classes with PSModel-dependent PRUController parameters, such
        as bsmp variable groups and so on.

    psdevstate.py :
        Implements PSDevState class, whose objects are used as mirrors
        to store device states.

"""



"""."""

from functools import partial as _part
from copy import deepcopy as _dcopy
import numpy as _np
from epics import PV as _PV

import mathphys.constants as _consts

from ...search import PSSearch as _PSS
from ...thread import RepeaterThread as _Repeater

from ..util import BaseClass as _BaseClass, ProcessImage as _ProcessImage

from .csdev import Const as _Const

C = _consts.light_speed
E0 = _consts.electron_rest_energy / _consts.elementary_charge * 1e-6  # in MeV


class CalcEnergy(_BaseClass, _Const):
    """."""

    def __init__(self, callback=None, dispersion=None, angle=None,
                 spectrometer=None):
        """."""
        super().__init__(callback=callback)
        self._dispersion = dispersion or self.DEFAULT_DISP
        self._angle = angle or self.DEFAULT_B_ANG
        self._spect = spectrometer or self.DEFAULT_SPECT
        self._excdata = _PSS.conv_psname_2_excdata(self._spect)
        self._currents = _np.array([], dtype=float)
        self._intdipole = _np.array([], dtype=float)
        self._energy = _np.array([], dtype=float)
        self._spread = _np.array([], dtype=float)
        self._beamcenter = _np.array([], dtype=float)
        self._beamsize = _np.array([], dtype=float)

    def get_map2write(self):
        """."""
        return {
            'Dispersion-SP': _part(self.write, 'dispersion'),
            'Angle-SP': _part(self.write, 'angle'),
            'Spectrometer-SP': _part(self.write, 'spectrometer'),
            }

    def get_map2read(self):
        """."""
        return {
            'Dispersion-RB': _part(self.read, 'dispersion'),
            'Angle-RB': _part(self.read, 'angle'),
            'Spectrometer-RB': _part(self.read, 'spectrometer'),
            'IntDipole-Mon': _part(self.read, 'intdipole'),
            'Energy-Mon': _part(self.read, 'energy'),
            'Spread-Mon': _part(self.read, 'spread'),
            }

    @property
    def dispersion(self):
        """."""
        return self._dispersion

    @dispersion.setter
    def dispersion(self, val):
        """."""
        if isinstance(val, (float, int)) and val != 0:
            self._dispersion = val
            self.run_callbacks('Dispersion-RB', val)
            self._perform_analysis()

    @property
    def angle(self):
        """."""
        return self._angle

    @angle.setter
    def angle(self, val):
        if isinstance(val, (float, int)) and val != 0 and abs(val) < _np.pi:
            self._angle = val
            self._perform_analysis()
            self.run_callbacks('Angle-RB', val)

    @property
    def spectrometer(self):
        """."""
        return self._spect

    @spectrometer.setter
    def spectrometer(self, val):
        """."""
        try:
            self._excdata = _PSS.conv_psname_2_excdata(val)
        except Exception:
            return
        self._spect = val
        self._perform_analysis()
        self.run_callbacks('Spectrometer-RB', val)

    @property
    def currents(self):
        """."""
        return _dcopy(self._currents)

    @property
    def beamsize(self):
        """."""
        return _dcopy(self._beamsize)

    @property
    def beamcenter(self):
        """."""
        return _dcopy(self._beamcenter)

    @property
    def intdipole(self):
        """."""
        return _dcopy(self._intdipole)

    @property
    def energy(self):
        """."""
        return _dcopy(self._energy)

    @property
    def spread(self):
        """."""
        return _dcopy(self._spread)

    def set_data(self, currents, beam_centers, beam_sizes):
        """."""
        if None in {currents, beam_centers, beam_sizes}:
            return False
        if isinstance(beam_sizes, (int, float)):
            beam_sizes = [beam_sizes]
        if isinstance(beam_centers, (int, float)):
            beam_centers = [beam_centers]
        if isinstance(currents, (int, float)):
            currents = [currents]
        beam_sizes = _np.array(beam_sizes, dtype=float)
        beam_centers = _np.array(beam_centers, dtype=float)
        currents = _np.array(currents, dtype=float)
        cond = beam_sizes.size != currents.size
        cond |= currents.size != beam_centers.size
        if cond:
            return False
        self._beamsize = beam_sizes
        self._beamcenter = beam_centers
        self._currents = currents
        self._perform_analysis()
        return True

    def _update_integrated_dipole(self):
        """."""
        if self._currents.size == 0:
            return
        mult = self._excdata.interp_curr2mult(
            currents=self._currents, only_main_harmonic=True)
        self._intdipole = mult['normal'][0]

    def _perform_analysis(self):
        """."""
        self._update_integrated_dipole()
        cond = self._intdipole.size == 0
        cond |= self._beamcenter.size == 0
        cond |= self._beamsize == 0
        if cond:
            return
        pc_nom = self._intdipole / self._angle * C * 1e-6  # in MeV
        pc = pc_nom * (1 + self._beamcenter / self._dispersion)
        self._energy = _np.sqrt(pc**2 + E0*E0)
        self._spread = self._beamsize / self._dispersion * 100  # in percent%
        self.run_callbacks('Energy-Mon', self._energy)
        self.run_callbacks('Spread-Mon', self._spread)
        self.run_callbacks('IntDipole-Mon', self._intdipole)


class MeasEnergy(_BaseClass, _Const):
    """."""

    def __init__(self, callback=None):
        """."""
        self.energy_calculator = CalcEnergy(callback=callback)
        self.image_processor = _ProcessImage(callback=callback)
        self.image_processor.imageflipx = self.image_processor.ImgFlip.On
        self.image_processor.imageflipy = self.image_processor.ImgFlip.Off
        self._profile = self.DEFAULT_PROFILE
        self._coefx = _PV(
            self.DEFAULT_PROFILE+':X:Gauss:Coef', callback=self._update_coefx)
        self._coefy = _PV(
            self.DEFAULT_PROFILE+':Y:Gauss:Coef', callback=self._update_coefy)
        self._width_source = _PV(self.DEFAULT_PROFILE + ':ROI:MaxSizeX_RBV')
        self._image_source = _PV(
            self.DEFAULT_PROFILE + ':RAW:ArrayData', auto_monitor=False)
        self._current_source = _PV(self.DEFAULT_SPECT + ':Current-Mon')
        super().__init__(callback=callback)
        self._thread = _Repeater(0.5, self.meas_energy, niter=0)
        self._thread.pause()
        self._thread.start()

    def get_map2write(self):
        """."""
        dic_ = self.image_processor.get_map2write()
        dic_.update(self.energy_calculator.get_map2write())
        dic_.update({'MeasureCtrl-Sel': _part(self.write, 'measuring')})
        return dic_

    def get_map2read(self):
        """."""
        dic_ = self.image_processor.get_map2read()
        dic_.update(self.energy_calculator.get_map2read())
        dic_.update({'MeasureCtrl-Sts': _part(self.read, 'measuring')})
        return dic_

    def start(self):
        """."""
        self._thread.resume()
        self.run_callbacks('MeasureCtrl-Sts', 1)

    def stop(self):
        """."""
        self._thread.pause()
        self.run_callbacks('MeasureCtrl-Sts', 0)

    @property
    def connected(self):
        """."""
        conn = self._coefx.connected
        conn &= self._coefy.connected
        conn &= self._width_source.connected
        conn &= self._image_source.connected
        return conn

    @property
    def current(self):
        """."""
        return self._current_source.get()

    @property
    def rate(self):
        """."""
        return 1/self._thread.interval

    @rate.setter
    def rate(self, val):
        """."""
        if isinstance(val, (float, int)) and 0 < val < 4:
            self._thread.interval = 1/val

    @property
    def measuring(self):
        """."""
        return not self._thread.is_paused()

    @measuring.setter
    def measuring(self, val):
        """."""
        if val:
            self.start()
        else:
            self.stop()

    def _update_coefx(self, pvname, value, **kwargs):
        """."""
        if value is None:
            return
        self.image_processor.px2mmscalex = value

    def _update_coefy(self, pvname, value, **kwargs):
        """."""
        if value is None:
            return
        self.image_processor.px2mmscaley = value

    def meas_energy(self):
        """."""
        value = self._width_source.value
        if isinstance(value, (float, int)):
            self.image_processor.imagewidth = int(value)

        self.image_processor.image = self._image_source.get()
        self.energy_calculator.set_data(
            self.current,
            self.image_processor.beamcentermmx,
            self.image_processor.beamsizemmx)
